# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2008  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Plugin framework.

The classes in this module make heavy use of Python container emulation. If
you are unfamiliar with this Python feature, see
http://docs.python.org/ref/sequence-types.html
"""
import logging
import operator
import re
import sys
import threading
import os
from os import path
import optparse  # pylint: disable=deprecated-module
import textwrap
import collections
import importlib

import six

from ipalib import errors
from ipalib.config import Env
from ipalib.text import _
from ipalib.util import classproperty
from ipalib.base import ReadOnly, lock, islocked
from ipalib.constants import DEFAULT_CONFIG
from ipapython import ipa_log_manager, ipautil
from ipapython.ipa_log_manager import (
    LOGGING_FORMAT_FILE,
    LOGGING_FORMAT_STDERR)
from ipapython.version import VERSION, API_VERSION, DEFAULT_PLUGINS

# pylint: disable=no-name-in-module, import-error
if six.PY3:
    from collections.abc import Mapping
else:
    from collections import Mapping
# pylint: enable=no-name-in-module, import-error

if six.PY3:
    unicode = str

logger = logging.getLogger(__name__)

# FIXME: Updated constants.TYPE_ERROR to use this clearer format from wehjit:
TYPE_ERROR = '%s: need a %r; got a %r: %r'


# FIXME: This function has no unit test
def find_modules_in_dir(src_dir):
    """
    Iterate through module names found in ``src_dir``.
    """
    if not (os.path.abspath(src_dir) == src_dir and os.path.isdir(src_dir)):
        return
    if os.path.islink(src_dir):
        return
    suffix = '.py'
    for name in sorted(os.listdir(src_dir)):
        if not name.endswith(suffix):
            continue
        pyfile = os.path.join(src_dir, name)
        if not os.path.isfile(pyfile):
            continue
        module = name[:-len(suffix)]
        if module == '__init__':
            continue
        yield module


class Registry:
    """A decorator that makes plugins available to the API

    Usage::

        register = Registry()

        @register()
        class obj_mod(...):
            ...

    For forward compatibility, make sure that the module-level instance of
    this object is named "register".
    """
    def __init__(self):
        self.__registry = collections.OrderedDict()

    def __call__(self, **kwargs):
        def register(plugin):
            """
            Register the plugin ``plugin``.

            :param plugin: A subclass of `Plugin` to attempt to register.
            """
            if not callable(plugin):
                raise TypeError('plugin must be callable; got %r' % plugin)

            # Raise DuplicateError if this exact class was already registered:
            if plugin in self.__registry:
                raise errors.PluginDuplicateError(plugin=plugin)

            # The plugin is okay, add to __registry:
            self.__registry[plugin] = dict(kwargs, plugin=plugin)

            return plugin

        return register

    def __iter__(self):
        return iter(self.__registry.values())


class Plugin(ReadOnly):
    """
    Base class for all plugins.
    """

    version = '1'

    def __init__(self, api):
        assert api is not None
        self.__api = api
        self.__finalize_called = False
        self.__finalized = False
        self.__finalize_lock = threading.RLock()

    @classmethod
    def __name_getter(cls):
        return cls.__name__

    # you know nothing, pylint
    name = classproperty(__name_getter)

    @classmethod
    def __full_name_getter(cls):
        return '{}/{}'.format(cls.name, cls.version)

    full_name = classproperty(__full_name_getter)

    @classmethod
    def __bases_getter(cls):
        return cls.__bases__

    bases = classproperty(__bases_getter)

    @classmethod
    def __doc_getter(cls):
        return cls.__doc__

    doc = classproperty(__doc_getter)

    @classmethod
    def __summary_getter(cls):
        doc = cls.doc
        if not _(doc).msg:
            return u'<%s.%s>' % (cls.__module__, cls.__name__)
        else:
            return unicode(doc).split('\n\n', 1)[0].strip()

    summary = classproperty(__summary_getter)

    @property
    def api(self):
        """
        Return `API` instance passed to `__init__()`.
        """
        return self.__api

    # FIXME: for backward compatibility only
    @property
    def env(self):
        return self.__api.env

    # FIXME: for backward compatibility only
    @property
    def Backend(self):
        return self.__api.Backend

    # FIXME: for backward compatibility only
    @property
    def Command(self):
        return self.__api.Command

    def finalize(self):
        """
        Finalize plugin initialization.

        This method calls `_on_finalize()` and locks the plugin object.

        Subclasses should not override this method. Custom finalization is done
        in `_on_finalize()`.
        """
        with self.__finalize_lock:
            assert self.__finalized is False
            if self.__finalize_called:
                # No recursive calls!
                return
            self.__finalize_called = True
            self._on_finalize()
            self.__finalized = True
            if not self.__api.is_production_mode():
                lock(self)

    def _on_finalize(self):
        """
        Do custom finalization.

        This method is called from `finalize()`. Subclasses can override this
        method in order to add custom finalization.
        """

    def ensure_finalized(self):
        """
        Finalize plugin initialization if it has not yet been finalized.
        """
        with self.__finalize_lock:
            if not self.__finalized:
                self.finalize()

    class finalize_attr:
        """
        Create a stub object for plugin attribute that isn't set until the
        finalization of the plugin initialization.

        When the stub object is accessed, it calls `ensure_finalized()` to make
        sure the plugin initialization is finalized. The stub object is expected
        to be replaced with the actual attribute value during the finalization
        (preferably in `_on_finalize()`), otherwise an `AttributeError` is
        raised.

        This is used to implement on-demand finalization of plugin
        initialization.
        """
        __slots__ = ('name', 'value')

        def __init__(self, name, value=None):
            self.name = name
            self.value = value

        def __get__(self, obj, cls):
            if obj is None or obj.api is None:
                return self.value
            obj.ensure_finalized()
            try:
                return getattr(obj, self.name)
            except RuntimeError:
                # If the actual attribute value is not set in _on_finalize(),
                # getattr() calls __get__() again, which leads to infinite
                # recursion. This can happen only if the plugin is written
                # badly, so advise the developer about that instead of giving
                # them a generic "maximum recursion depth exceeded" error.
                raise AttributeError(
                    "attribute '%s' of plugin '%s' was not set in finalize()" % (self.name, obj.name)
                )

    def __repr__(self):
        """
        Return 'module_name.class_name()' representation.

        This representation could be used to instantiate this Plugin
        instance given the appropriate environment.
        """
        return '%s.%s()' % (
            self.__class__.__module__,
            self.__class__.__name__
        )


class APINameSpace(Mapping):
    def __init__(self, api, base):
        self.__api = api
        self.__base = base
        self.__plugins = None
        self.__plugins_by_key = None

    def __enumerate(self):
        if self.__plugins is not None and self.__plugins_by_key is not None:
            return

        default_map = self.__api._API__default_map
        plugins = set()
        key_dict = self.__plugins_by_key = {}

        for plugin in self.__api._API__plugins:
            if not any(issubclass(b, self.__base) for b in plugin.bases):
                continue
            plugins.add(plugin)
            key_dict[plugin] = plugin
            key_dict[plugin.name, plugin.version] = plugin
            key_dict[plugin.full_name] = plugin
            if plugin.version == default_map.get(plugin.name, '1'):
                key_dict[plugin.name] = plugin

        self.__plugins = sorted(plugins, key=operator.attrgetter('full_name'))

    def __len__(self):
        self.__enumerate()
        return len(self.__plugins)

    def __contains__(self, key):
        self.__enumerate()
        return key in self.__plugins_by_key

    def __iter__(self):
        self.__enumerate()
        return iter(self.__plugins)

    def get_plugin(self, key):
        self.__enumerate()
        return self.__plugins_by_key[key]

    def __getitem__(self, key):
        plugin = self.get_plugin(key)
        return self.__api._get(plugin)

    def __call__(self):
        return six.itervalues(self)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)


class API(ReadOnly):
    """
    Dynamic API object through which `Plugin` instances are accessed.
    """

    def __init__(self):
        super(API, self).__init__()
        self.__plugins = set()
        self.__plugins_by_key = {}
        self.__default_map = {}
        self.__instances = {}
        self.__next = {}
        self.__done = set()
        self.env = Env()

    @property
    def bases(self):
        raise NotImplementedError

    @property
    def packages(self):
        raise NotImplementedError

    def __len__(self):
        """
        Return the number of plugin namespaces in this API object.
        """
        return len(self.bases)

    def __iter__(self):
        """
        Iterate (in ascending order) through plugin namespace names.
        """
        return (base.__name__ for base in self.bases)

    def __contains__(self, name):
        """
        Return True if this API object contains plugin namespace ``name``.

        :param name: The plugin namespace name to test for membership.
        """
        return name in set(self)

    def __getitem__(self, name):
        """
        Return the plugin namespace corresponding to ``name``.

        :param name: The name of the plugin namespace you wish to retrieve.
        """
        if name in self:
            try:
                return getattr(self, name)
            except AttributeError:
                pass

        raise KeyError(name)

    def __call__(self):
        """
        Iterate (in ascending order by name) through plugin namespaces.
        """
        for name in self:
            try:
                yield getattr(self, name)
            except AttributeError:
                raise KeyError(name)

    def is_production_mode(self):
        """
        If the object has self.env.mode defined and that mode is
        production return True, otherwise return False.
        """
        return getattr(self.env, 'mode', None) == 'production'

    def __doing(self, name):
        if name in self.__done:
            raise Exception(
                '%s.%s() already called' % (self.__class__.__name__, name)
            )
        self.__done.add(name)

    def __do_if_not_done(self, name):
        if name not in self.__done:
            getattr(self, name)()

    def isdone(self, name):
        return name in self.__done

    def bootstrap(self, parser=None, **overrides):
        """
        Initialize environment variables and logging.
        """
        self.__doing('bootstrap')
        self.env._bootstrap(**overrides)
        self.env._finalize_core(**dict(DEFAULT_CONFIG))

        # Add the argument parser
        if not parser:
            parser = self.build_global_parser()
        self.parser = parser

        root_logger = logging.getLogger()

        # If logging has already been configured somewhere else (like in the
        # installer), don't add handlers or change levels:
        if root_logger.handlers or self.env.validate_api:
            return

        if self.env.debug:
            level = logging.DEBUG
        else:
            level = logging.INFO
        root_logger.setLevel(level)

        for attr in self.env:
            match = re.match(r'^log_logger_level_'
                             r'(debug|info|warn|warning|error|critical|\d+)$',
                             attr)
            if not match:
                continue

            level = ipa_log_manager.convert_log_level(match.group(1))

            value = getattr(self.env, attr)
            regexps = re.split('\s*,\s*', value)

            # Add the regexp, it maps to the configured level
            for regexp in regexps:
                root_logger.addFilter(ipa_log_manager.Filter(regexp, level))

        # Add stderr handler:
        level = logging.INFO
        if self.env.debug:
            level = logging.DEBUG
        else:
            if self.env.context == 'cli':
                if self.env.verbose > 0:
                    level = logging.INFO
                else:
                    level = logging.WARNING
        handler = logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(ipa_log_manager.Formatter(LOGGING_FORMAT_STDERR))
        root_logger.addHandler(handler)

        # check after logging is set up but before we create files.
        fse = sys.getfilesystemencoding()
        if fse.lower() not in {'utf-8', 'utf8'}:
            raise errors.SystemEncodingError(encoding=fse)

        # Add file handler:
        if self.env.mode in ('dummy', 'unit_test'):
            return  # But not if in unit-test mode
        if self.env.log is None:
            return
        log_dir = path.dirname(self.env.log)
        if not path.isdir(log_dir):
            try:
                os.makedirs(log_dir)
            except OSError:
                logger.error('Could not create log_dir %r', log_dir)
                return

        level = logging.INFO
        if self.env.debug:
            level = logging.DEBUG
        try:
            handler = logging.FileHandler(self.env.log)
        except IOError as e:
            logger.error('Cannot open log file %r: %s', self.env.log, e)
            return
        handler.setLevel(level)
        handler.setFormatter(ipa_log_manager.Formatter(LOGGING_FORMAT_FILE))
        root_logger.addHandler(handler)

    def build_global_parser(self, parser=None, context=None):
        """
        Add global options to an optparse.OptionParser instance.
        """
        def config_file_callback(option, opt, value, parser):
            if not os.path.isfile(value):
                parser.error(
                    _("%(filename)s: file not found") % dict(filename=value))

            parser.values.conf = value

        if parser is None:
            parser = optparse.OptionParser(
                add_help_option=False,
                formatter=IPAHelpFormatter(),
                usage='%prog [global-options] COMMAND [command-options]',
                description='Manage an IPA domain',
                version=('VERSION: %s, API_VERSION: %s' %
                            (VERSION, API_VERSION)),
                epilog='\n'.join([
                    'See "ipa help topics" for available help topics.',
                    'See "ipa help <TOPIC>" for more information on '
                    + 'a specific topic.',
                    'See "ipa help commands" for the full list of commands.',
                    'See "ipa <COMMAND> --help" for more information on '
                    + 'a specific command.',
                ]))
            parser.disable_interspersed_args()
            parser.add_option("-h", "--help", action="help",
                help='Show this help message and exit')

        parser.add_option('-e', dest='env', metavar='KEY=VAL', action='append',
            help='Set environment variable KEY to VAL',
        )
        parser.add_option('-c', dest='conf', metavar='FILE', action='callback',
            callback=config_file_callback, type='string',
            help='Load configuration from FILE.',
        )
        parser.add_option('-d', '--debug', action='store_true',
            help='Produce full debuging output',
        )
        parser.add_option('--delegate', action='store_true',
            help='Delegate the TGT to the IPA server',
        )
        parser.add_option('-v', '--verbose', action='count',
            help='Produce more verbose output. A second -v displays the XML-RPC request',
        )
        if context == 'cli':
            parser.add_option('-a', '--prompt-all', action='store_true',
                help='Prompt for ALL values (even if optional)'
            )
            parser.add_option('-n', '--no-prompt', action='store_false',
                dest='interactive',
                help='Prompt for NO values (even if required)'
            )
            parser.add_option('-f', '--no-fallback', action='store_false',
                dest='fallback',
                help='Only use the server configured in /etc/ipa/default.conf'
            )

        return parser

    def bootstrap_with_global_options(self, parser=None, context=None):
        parser = self.build_global_parser(parser, context)
        (options, args) = parser.parse_args()
        overrides = {}
        if options.env is not None:
            assert type(options.env) is list
            for item in options.env:
                try:
                    (key, value) = item.split('=', 1)
                except ValueError:
                    # FIXME: this should raise an IPA exception with an
                    # error code.
                    # --Jason, 2008-10-31
                    pass
                overrides[str(key.strip())] = value.strip()
        for key in ('conf', 'debug', 'verbose', 'prompt_all', 'interactive',
            'fallback', 'delegate'):
            value = getattr(options, key, None)
            if value is not None:
                overrides[key] = value
        if hasattr(options, 'prod'):
            overrides['webui_prod'] = options.prod
        if context is not None:
            overrides['context'] = context
        self.bootstrap(parser, **overrides)
        return (options, args)

    def load_plugins(self):
        """
        Load plugins from all standard locations.

        `API.bootstrap` will automatically be called if it hasn't been
        already.
        """
        self.__doing('load_plugins')
        self.__do_if_not_done('bootstrap')
        if self.env.mode in ('dummy', 'unit_test'):
            return
        for package in self.packages:
            self.add_package(package)

    # FIXME: This method has no unit test
    def add_package(self, package):
        """
        Add plugin modules from the ``package``.

        :param package: A package from which to add modules.
        """
        package_name = package.__name__
        package_file = package.__file__
        package_dir = path.dirname(path.abspath(package_file))

        parent = sys.modules[package_name.rpartition('.')[0]]
        parent_dir = path.dirname(path.abspath(parent.__file__))
        if parent_dir == package_dir:
            raise errors.PluginsPackageError(
                name=package_name, file=package_file
            )

        logger.debug("importing all plugin modules in %s...", package_name)
        modules = getattr(package, 'modules', find_modules_in_dir(package_dir))
        modules = ['.'.join((package_name, mname)) for mname in modules]

        for name in modules:
            logger.debug("importing plugin module %s", name)
            try:
                module = importlib.import_module(name)
            except errors.SkipPluginModule as e:
                logger.debug("skipping plugin module %s: %s", name, e.reason)
                continue
            except Exception as e:
                if self.env.startup_traceback:
                    logger.exception("could not load plugin module %s", name)
                raise

            try:
                self.add_module(module)
            except errors.PluginModuleError as e:
                logger.debug("%s", e)

    def add_module(self, module):
        """
        Add plugins from the ``module``.

        :param module: A module from which to add plugins.
        """
        try:
            register = module.register
        except AttributeError:
            pass
        else:
            if isinstance(register, Registry):
                for kwargs in register:
                    self.add_plugin(**kwargs)
                return

        raise errors.PluginModuleError(name=module.__name__)

    def add_plugin(self, plugin, override=False, no_fail=False):
        """
        Add the plugin ``plugin``.

        :param plugin: A subclass of `Plugin` to attempt to add.
        :param override: If true, override an already added plugin.
        """
        if not callable(plugin):
            raise TypeError('plugin must be callable; got %r' % plugin)

        # Find the base class or raise SubclassError:
        for base in plugin.bases:
            if issubclass(base, self.bases):
                break
        else:
            raise errors.PluginSubclassError(
                plugin=plugin,
                bases=self.bases,
            )

        # Check override:
        prev = self.__plugins_by_key.get(plugin.full_name)
        if prev:
            if not override:
                if no_fail:
                    return
                else:
                    # Must use override=True to override:
                    raise errors.PluginOverrideError(
                        base=base.__name__,
                        name=plugin.name,
                        plugin=plugin,
                    )

            self.__plugins.remove(prev)
            self.__next[plugin] = prev
        else:
            if override:
                if no_fail:
                    return
                else:
                    # There was nothing already registered to override:
                    raise errors.PluginMissingOverrideError(
                        base=base.__name__,
                        name=plugin.name,
                        plugin=plugin,
                    )

        # The plugin is okay, add to sub_d:
        self.__plugins.add(plugin)
        self.__plugins_by_key[plugin.full_name] = plugin

    def finalize(self):
        """
        Finalize the registration, instantiate the plugins.

        `API.bootstrap` will automatically be called if it hasn't been
        already.
        """
        self.__doing('finalize')
        self.__do_if_not_done('load_plugins')

        if self.env.env_confdir is not None:
            if self.env.env_confdir == self.env.confdir:
                logger.info(
                    "IPA_CONFDIR env sets confdir to '%s'.", self.env.confdir)

        for plugin in self.__plugins:
            if not self.env.validate_api:
                if plugin.full_name not in DEFAULT_PLUGINS:
                    continue
            else:
                try:
                    default_version = self.__default_map[plugin.name]
                except KeyError:
                    pass
                else:
                    # Technicall plugin.version is not an API version. The
                    # APIVersion class can handle plugin versions. It's more
                    # lean than pkg_resource.parse_version().
                    version = ipautil.APIVersion(plugin.version)
                    default_version = ipautil.APIVersion(default_version)
                    if version < default_version:
                        continue
            self.__default_map[plugin.name] = plugin.version

        production_mode = self.is_production_mode()

        for base in self.bases:
            for plugin in self.__plugins:
                if not any(issubclass(b, base) for b in plugin.bases):
                    continue
                if not self.env.plugins_on_demand:
                    self._get(plugin)

            name = base.__name__
            if not production_mode:
                assert not hasattr(self, name)
            setattr(self, name, APINameSpace(self, base))

        for instance in six.itervalues(self.__instances):
            if not production_mode:
                assert instance.api is self
            if not self.env.plugins_on_demand:
                instance.ensure_finalized()
                if not production_mode:
                    assert islocked(instance)

        self.__finalized = True

        if not production_mode:
            lock(self)

    def _get(self, plugin):
        if not callable(plugin):
            raise TypeError('plugin must be callable; got %r' % plugin)
        if plugin not in self.__plugins:
            raise KeyError(plugin)

        try:
            instance = self.__instances[plugin]
        except KeyError:
            instance = self.__instances[plugin] = plugin(self)

        return instance

    def get_plugin_next(self, plugin):
        if not callable(plugin):
            raise TypeError('plugin must be callable; got %r' % plugin)

        return self.__next[plugin]


class IPAHelpFormatter(optparse.IndentedHelpFormatter):
    def format_epilog(self, text):
        text_width = self.width - self.current_indent
        indent = " " * self.current_indent
        lines = text.splitlines()
        result = '\n'.join(
            textwrap.fill(line, text_width, initial_indent=indent,
                subsequent_indent=indent)
            for line in lines)
        return '\n%s\n' % result



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

from __future__ import absolute_import

from ipalib import (
    _,
    ngettext,
    api,
    DNSNameParam,
    Str,
    StrEnum,
    errors,
)
from ipalib.frontend import Local
from ipalib.plugable import Registry
from ipalib.util import (
    normalize_hostname,
    hostname_validator,
    validate_bind_forwarder,
)
from ipaserver.plugins.baseldap import (
    LDAPObject,
    LDAPRetrieve,
    LDAPUpdate,
    LDAPSearch,
    LDAPCreate,
    LDAPDelete,
)
from .dns import dns_container_exists
from ipapython.dn import DN


__doc__ = _("""
DNS server configuration
""") + _("""
Manipulate DNS server configuration
""") + _("""
EXAMPLES:
""") + _("""
  Show configuration of a specific DNS server:
    ipa dnsserver-show
""") + _("""
  Update configuration of a specific DNS server:
    ipa dnsserver-mod
""")


register = Registry()

topic = None

dnsserver_object_class = ['top', 'idnsServerConfigObject']

@register()
class dnsserver(LDAPObject):
    """
    DNS Servers
    """
    container_dn = api.env.container_dnsservers
    object_name = _('DNS server')
    object_name_plural = _('DNS servers')
    object_class = dnsserver_object_class
    default_attributes = [
        'idnsServerId',
        'idnsSOAmName',
        'idnsForwarders',
        'idnsForwardPolicy',
    ]
    label = _('DNS Servers')
    label_singular = _('DNS Server')

    permission_filter_objectclasses = ['idnsServerConfigObject']

    managed_permissions = {
        'System: Read DNS Servers Configuration': {
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'objectclass',
                'idnsServerId',
                'idnsSOAmName',
                'idnsForwarders',
                'idnsForwardPolicy',
                'idnsSubstitutionVariable',
            },
            'ipapermlocation': api.env.basedn,
            'default_privileges': {
                'DNS Servers',
                'DNS Administrators'
            },
        },
        'System: Modify DNS Servers Configuration': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {
                'idnsSOAmName',
                'idnsForwarders',
                'idnsForwardPolicy',
                'idnsSubstitutionVariable',
            },
            'ipapermlocation': api.env.basedn,
            'default_privileges': {'DNS Administrators'},
        },
    }

    takes_params = (
        Str(
            'idnsserverid',
            hostname_validator,
            cli_name='hostname',
            primary_key=True,
            label=_('Server name'),
            doc=_('DNS Server name'),
            normalizer=normalize_hostname,
        ),
        DNSNameParam(
            'idnssoamname?',
            cli_name='soa_mname_override',
            label=_('SOA mname override'),
            doc=_('SOA mname (authoritative server) override'),
        ),
        Str(
            'idnsforwarders*',
            validate_bind_forwarder,
            cli_name='forwarder',
            label=_('Forwarders'),
            doc=_(
                'Per-server forwarders. A custom port can be specified '
                'for each forwarder using a standard format '
                '"IP_ADDRESS port PORT"'
            ),
        ),
        StrEnum(
            'idnsforwardpolicy?',
            cli_name='forward_policy',
            label=_('Forward policy'),
            doc=_(
                'Per-server conditional forwarding policy. Set to "none" to '
                'disable forwarding to global forwarder for this zone. In '
                'that case, conditional zone forwarders are disregarded.'
            ),
            values=(u'only', u'first', u'none'),
        ),
    )

    def get_dn(self, *keys, **options):
        if not dns_container_exists(self.api.Backend.ldap2):
            raise errors.NotFound(reason=_('DNS is not configured'))
        return super(dnsserver, self).get_dn(*keys, **options)


@register()
class dnsserver_mod(LDAPUpdate):
    __doc__ = _('Modify DNS server configuration')
    topic = 'dns'

    msg_summary = _('Modified DNS server "%(value)s"')


@register()
class dnsserver_find(LDAPSearch):
    __doc__ = _('Search for DNS servers.')
    topic = 'dns'

    msg_summary = ngettext(
        '%(count)d DNS server matched',
        '%(count)d DNS servers matched', 0
    )

    def pre_callback(self, ldap, filters, attrs_list,
                     base_dn, scope, *args, **options):
        assert isinstance(base_dn, DN)

        if not dns_container_exists(self.api.Backend.ldap2):
            raise errors.InvocationError(
                format=_('IPA DNS Server is not installed'))

        return (filters, base_dn, scope)


@register()
class dnsserver_show(LDAPRetrieve):
    __doc__=_('Display configuration of a DNS server.')
    topic = 'dns'


@register()
class dnsserver_add(LDAPCreate, Local):
    """
    Only for internal use, this is not part of public API on purpose.
    Be careful in future this will be transformed to public API call
    """
    __doc__ = _('Add a new DNS server.')
    topic = 'dns'

    msg_summary = _('Added new DNS server "%(value)s"')


@register()
class dnsserver_del(LDAPDelete, Local):
    """
    Only for internal use, this is not part of public API on purpose.
    Be careful in future this will be transformed to public API call
    """
    __doc__ = _('Delete a DNS server')
    topic = 'dns'

    msg_summary = _('Deleted DNS server "%(value)s"')



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

from __future__ import absolute_import

import logging

from ipalib.plugable import Registry
from ipalib import errors
from ipalib import Updater
from ipaplatform.paths import paths
from ipapython.dn import DN
from ipaserver.install import sysupgrade
from ipaserver.install.ldapupdate import LDAPUpdate

logger = logging.getLogger(__name__)

register = Registry()


@register()
class update_nis_configuration(Updater):
    """Update NIS configuration

    NIS configuration can be updated only if NIS Server was configured via
    ipa-nis-manage command.
    """

    def __recover_from_missing_maps(self, ldap):
        # https://fedorahosted.org/freeipa/ticket/5507
        # if all following DNs are missing, but 'NIS Server' container exists
        # we are experiencig bug and maps should be fixed

        if sysupgrade.get_upgrade_state('nis',
                                        'done_recover_from_missing_maps'):
            # this recover must be done only once, a user may deleted some
            # maps, we do not want to restore them again
            return

        logger.debug("Recovering from missing NIS maps bug")

        suffix = "cn=NIS Server,cn=plugins,cn=config"
        domain = self.api.env.domain
        missing_dn_list = [
            DN(nis_map.format(domain=domain, suffix=suffix)) for nis_map in [
                "nis-domain={domain}+nis-map=passwd.byname,{suffix}",
                "nis-domain={domain}+nis-map=passwd.byuid,{suffix}",
                "nis-domain={domain}+nis-map=group.byname,{suffix}",
                "nis-domain={domain}+nis-map=group.bygid,{suffix}",
                "nis-domain={domain}+nis-map=netid.byname,{suffix}",
                "nis-domain={domain}+nis-map=netgroup,{suffix}",
            ]
        ]

        for dn in missing_dn_list:
            try:
                ldap.get_entry(dn, attrs_list=['cn'])
            except errors.NotFound:
                pass
            else:
                # bug is not effective, at least one of 'possible missing'
                # maps was detected
                return

        sysupgrade.set_upgrade_state('nis', 'done_recover_from_missing_maps',
                                     True)

        # bug is effective run update to recreate missing maps
        ld = LDAPUpdate(sub_dict={}, ldapi=True)
        ld.update([paths.NIS_ULDIF])

    def execute(self, **options):
        ldap = self.api.Backend.ldap2
        dn = DN(('cn', 'NIS Server'), ('cn', 'plugins'), ('cn', 'config'))
        try:
            ldap.get_entry(dn, attrs_list=['cn'])
        except errors.NotFound:
            # NIS is not configured on system, do not execute update
            logger.debug("Skipping NIS update, NIS Server is not configured")

            # container does not exist, bug #5507 is not effective
            sysupgrade.set_upgrade_state(
                'nis', 'done_recover_from_missing_maps', True)
        else:
            self.__recover_from_missing_maps(ldap)

            logger.debug("Executing NIS Server update")
            ld = LDAPUpdate(sub_dict={}, ldapi=True)
            ld.update([paths.NIS_UPDATE_ULDIF])

        return False, ()



# Authors:
#   Andrew Wnuk <awnuk@redhat.com>
#   Jason Gerard DeRose <jderose@redhat.com>
#   John Dennis <jdennis@redhat.com>
#
# Copyright (C) 2009  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64

import six

from ipaclient.frontend import MethodOverride
from ipalib import errors
from ipalib import x509
from ipalib import util
from ipalib.parameters import BinaryFile, File, Flag, Str
from ipalib.plugable import Registry
from ipalib.text import _

if six.PY3:
    unicode = str

register = Registry()


class CertRetrieveOverride(MethodOverride):
    takes_options = (
        Str(
            'certificate_out?',
            doc=_('Write certificate (chain if --chain used) to file'),
            include='cli',
            cli_metavar='FILE',
        ),
    )

    def forward(self, *args, **options):
        if 'certificate_out' in options:
            certificate_out = options.pop('certificate_out')
            try:
                util.check_writable_file(certificate_out)
            except errors.FileError as e:
                raise errors.ValidationError(name='certificate-out',
                                             error=str(e))
        else:
            certificate_out = None

        result = super(CertRetrieveOverride, self).forward(*args, **options)

        if certificate_out is not None:
            if options.get('chain', False):
                certs = result['result']['certificate_chain']
            else:
                certs = [base64.b64decode(result['result']['certificate'])]
            certs = (x509.load_der_x509_certificate(cert) for cert in certs)
            x509.write_certificate_list(certs, certificate_out)

        return result


@register(override=True, no_fail=True)
class cert_request(CertRetrieveOverride):
    takes_options = CertRetrieveOverride.takes_options + (
        Str(
            'database?',
            label=_('Path to NSS database'),
            doc=_('Path to NSS database to use for private key'),
        ),
        Str(
            'private_key?',
            label=_('Path to private key file'),
            doc=_('Path to PEM file containing a private key'),
        ),
        Str(
            'password_file?',
            label=_(
                'File containing a password for the private key or database'),
        ),
        Str(
            'csr_profile_id?',
            label=_('Name of CSR generation profile (if not the same as'
                    ' profile_id)'),
        ),
    )

    def get_args(self):
        for arg in super(cert_request, self).get_args():
            if arg.name == 'csr':
                arg = arg.clone_retype(arg.name, File, required=False)
            yield arg

    def forward(self, csr=None, **options):
        database = options.pop('database', None)
        private_key = options.pop('private_key', None)
        csr_profile_id = options.pop('csr_profile_id', None)
        password_file = options.pop('password_file', None)

        if csr is None:
            # Deferred import, ipaclient.csrgen is expensive to load.
            # see https://pagure.io/freeipa/issue/7484
            from ipaclient import csrgen

            if database:
                adaptor = csrgen.NSSAdaptor(database, password_file)
            elif private_key:
                adaptor = csrgen.OpenSSLAdaptor(
                    key_filename=private_key, password_filename=password_file)
            else:
                raise errors.InvocationError(
                    message=u"One of 'database' or 'private_key' is required")

            pubkey_info = adaptor.get_subject_public_key_info()
            pubkey_info_b64 = base64.b64encode(pubkey_info)

            # If csr_profile_id is passed, that takes precedence.
            # Otherwise, use profile_id. If neither are passed, the default
            # in cert_get_requestdata will be used.
            profile_id = csr_profile_id
            if profile_id is None:
                profile_id = options.get('profile_id')

            response = self.api.Command.cert_get_requestdata(
                profile_id=profile_id,
                principal=options.get('principal'),
                public_key_info=pubkey_info_b64)

            req_info_b64 = response['result']['request_info']
            req_info = base64.b64decode(req_info_b64)

            csr = adaptor.sign_csr(req_info)

            if not csr:
                raise errors.CertificateOperationError(
                    error=(_('Generated CSR was empty')))

        else:
            if database is not None or private_key is not None:
                raise errors.MutuallyExclusiveError(reason=_(
                    "Options 'database' and 'private_key' are not compatible"
                    " with 'csr'"))

        return super(cert_request, self).forward(csr, **options)


@register(override=True, no_fail=True)
class cert_show(CertRetrieveOverride):
    def get_options(self):
        for option in super(cert_show, self).get_options():
            if option.name == 'out':
                # skip server-defined --out
                continue
            if option.name == 'certificate_out':
                # add --out as a deprecated alias of --certificate-out
                option = option.clone_rename(
                    'out',
                    cli_name='certificate_out',
                    deprecated_cli_aliases={'out'},
                )
            yield option

    def forward(self, *args, **options):
        try:
            options['certificate_out'] = options.pop('out')
        except KeyError:
            pass

        return super(cert_show, self).forward(*args, **options)


@register(override=True, no_fail=True)
class cert_remove_hold(MethodOverride):
    has_output_params = (
        Flag('unrevoked',
            label=_('Unrevoked'),
        ),
        Str('error_string',
            label=_('Error'),
        ),
    )


@register(override=True, no_fail=True)
class cert_find(MethodOverride):
    takes_options = (
        BinaryFile(
            'file?',
            label=_("Input filename"),
            doc=_('File to load the certificate from.'),
            include='cli',
        ),
    )

    def forward(self, *args, **options):
        if self.api.env.context == 'cli':
            if 'certificate' in options and 'file' in options:
                raise errors.MutuallyExclusiveError(
                    reason=_("cannot specify both raw certificate and file"))
            if 'certificate' not in options and 'file' in options:
                options['certificate'] = x509.load_unknown_x509_certificate(
                                            options.pop('file'))

        return super(cert_find, self).forward(*args, **options)



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#
from __future__ import absolute_import

import six

from ipapython.dn import DN
from ipatests.test_xmlrpc.tracker.base import Tracker
from ipatests.util import assert_deepequal
from ipatests.test_xmlrpc.xmlrpc_test import (
    fuzzy_issuer,
    fuzzy_caid,
    fuzzy_base64,
    fuzzy_sequence_of,
    fuzzy_bytes,
)
from ipatests.test_xmlrpc import objectclasses


if six.PY3:
    unicode = str


class CATracker(Tracker):
    """Implementation of a Tracker class for CA plugin."""

    ldap_keys = {
        'dn', 'cn', 'ipacaid', 'ipacasubjectdn', 'ipacaissuerdn', 'description'
    }
    cert_keys = {
        'certificate',
    }
    cert_all_keys = {
        'certificate_chain',
    }
    find_keys = ldap_keys
    find_all_keys = {'objectclass'} | ldap_keys
    retrieve_keys = ldap_keys | cert_keys
    retrieve_all_keys = {'objectclass'} | retrieve_keys | cert_all_keys
    create_keys = {'objectclass'} | retrieve_keys
    update_keys = ldap_keys - {'dn'}

    def __init__(self, name, subject, desc=u"Test generated CA",
                 default_version=None):
        super(CATracker, self).__init__(default_version=default_version)
        self.attrs = {}
        self.ipasubjectdn = subject
        self.description = desc

        self.dn = DN(('cn', name),
                     self.api.env.container_ca,
                     self.api.env.basedn)

    def make_create_command(self):
        """Make function that creates the plugin entry object."""
        return self.make_command(
            'ca_add', self.name, ipacasubjectdn=self.ipasubjectdn,
            description=self.description
        )

    def check_create(self, result):
        assert_deepequal(dict(
            value=self.name,
            summary=u'Created CA "{}"'.format(self.name),
            result=dict(self.filter_attrs(self.create_keys))
        ), result)

    def track_create(self):
        self.attrs = dict(
            dn=unicode(self.dn),
            cn=[self.name],
            description=[self.description],
            ipacasubjectdn=[self.ipasubjectdn],
            ipacaissuerdn=[fuzzy_issuer],
            ipacaid=[fuzzy_caid],
            certificate=fuzzy_base64,
            certificate_chain=fuzzy_sequence_of(fuzzy_bytes),
            objectclass=objectclasses.ca
        )
        self.exists = True

    def make_delete_command(self):
        """Make function that deletes the plugin entry object."""
        return self.make_command('ca_del', self.name)

    def check_delete(self, result):
        assert_deepequal(dict(
            value=[self.name],
            summary=u'Deleted CA "{}"'.format(self.name),
            result=dict(failed=[])
        ), result)

    def make_retrieve_command(self, all=False, raw=False, **options):
        """Make function that retrieves the entry using ${CMD}_show"""
        return self.make_command('ca_show', self.name, all=all, raw=raw,
                                 **options)

    def check_retrieve(self, result, all=False, raw=False):
        """Check the plugin's `show` command result"""
        if all:
            expected = self.filter_attrs(self.retrieve_all_keys)
        else:
            expected = self.filter_attrs(self.retrieve_keys)

        assert_deepequal(dict(
            value=self.name,
            summary=None,
            result=expected
        ), result)

    def make_find_command(self, *args, **kwargs):
        """Make function that finds the entry using ${CMD}_find

        Note that the name (or other search terms) needs to be specified
        in arguments.
        """
        return self.make_command('ca_find', *args, **kwargs)

    def check_find(self, result, all=False, raw=False):
        """Check the plugin's `find` command result"""
        if all:
            expected = self.filter_attrs(self.find_all_keys)
        else:
            expected = self.filter_attrs(self.find_keys)

        assert_deepequal(dict(
            count=1,
            truncated=False,
            summary=u'1 CA matched',
            result=[expected]
        ), result)

    def make_update_command(self, updates):
        """Make function that modifies the entry using ${CMD}_mod"""
        return self.make_command('ca_mod', self.name, **updates)

    def check_update(self, result, extra_keys=()):
        """Check the plugin's `find` command result"""
        assert_deepequal(dict(
            value=self.name,
            summary=u'Modified CA "{}"'.format(self.name),
            result=self.filter_attrs(self.update_keys | set(extra_keys))
        ), result)



# Authors:
#     Alexander Bokovoy <abokovoy@redhat.com>
#
# Copyright (C) 2011-2016  Red Hat
# see file 'COPYING' for use and warranty information
#
# Portions (C) Andrew Tridgell, Andrew Bartlett
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Make sure we only run this module at the server where samba4-python
# package is installed to avoid issues with unavailable modules

from __future__ import absolute_import

import logging
import re
import time

from ipalib import api, _
from ipalib import errors
from ipapython import ipautil
from ipapython.dn import DN
from ipapython.dnsutil import query_srv
from ipapython.ipaldap import ldap_initialize
from ipaserver.install import installutils
from ipaserver.dcerpc_common import (TRUST_BIDIRECTIONAL,
                                     TRUST_JOIN_EXTERNAL,
                                     trust_type_string)

from ipalib.util import normalize_name

import os
import struct
import random

from samba import param
from samba import credentials
from samba.dcerpc import security, lsa, drsblobs, nbt, netlogon
from samba.ndr import ndr_pack, ndr_print
from samba import net
from samba import arcfour_encrypt
from samba import ntstatus
import samba

import ldap as _ldap
from ipapython import ipaldap
from ipapython.dnsutil import DNSName
from dns.exception import DNSException
import pysss_nss_idmap
import pysss
import six
from ipaplatform.paths import paths

from ldap.filter import escape_filter_chars
from time import sleep

try:
    from ldap.controls import RequestControl as LDAPControl
except ImportError:
    from ldap.controls import LDAPControl

if six.PY3:
    unicode = str
    long = int

__doc__ = _("""
Classes to manage trust joins using DCE-RPC calls

The code in this module relies heavily on samba4-python package
and Samba4 python bindings.
""")

logger = logging.getLogger(__name__)


def is_sid_valid(sid):
    try:
        security.dom_sid(sid)
    except TypeError:
        return False
    else:
        return True


access_denied_error = errors.ACIError(
                          info=_('CIFS server denied your credentials'))
dcerpc_error_codes = {
    -1073741823:
        errors.RemoteRetrieveError(
            reason=_('communication with CIFS server was unsuccessful')),
    -1073741790: access_denied_error,
    -1073741715: access_denied_error,
    -1073741614: access_denied_error,
    -1073741603:
        errors.ValidationError(
            name=_('AD domain controller'),
            error=_('unsupported functional level')),
    -1073741811:  # NT_STATUS_INVALID_PARAMETER
        errors.RemoteRetrieveError(
            reason=_('AD domain controller complains about communication '
                     'sequence. It may mean unsynchronized time on both '
                     'sides, for example')),
    -1073741776:  # NT_STATUS_INVALID_PARAMETER_MIX,
                  # we simply will skip the binding
        access_denied_error,
    -1073741772:  # NT_STATUS_OBJECT_NAME_NOT_FOUND
        errors.NotFound(
            reason=_('Cannot find specified domain or server name')),
}

dcerpc_error_messages = {
    "NT_STATUS_OBJECT_NAME_NOT_FOUND":
        errors.NotFound(
            reason=_('Cannot find specified domain or server name')),
    "The object name is not found.":
        errors.NotFound(
            reason=_('Cannot find specified domain or server name')),
    "WERR_NO_LOGON_SERVERS":
        errors.RemoteRetrieveError(
            reason=_('AD DC was unable to reach any IPA domain controller. '
                     'Most likely it is a DNS or firewall issue')),
    # This is a very long key, don't change it
    "There are currently no logon servers available to "
    "service the logon request.":
        errors.RemoteRetrieveError(
            reason=_('AD DC was unable to reach any IPA domain controller. '
                     'Most likely it is a DNS or firewall issue')),
    "NT_STATUS_INVALID_PARAMETER_MIX":
        errors.RequirementError(
            name=_('At least the domain or IP address should be specified')),
}

pysss_type_key_translation_dict = {
    pysss_nss_idmap.ID_USER: 'user',
    pysss_nss_idmap.ID_GROUP: 'group',
    # Used for users with magic private groups
    pysss_nss_idmap.ID_BOTH: 'both',
}


class TrustTopologyConflictSolved(Exception):
    """
    Internal trust error: raised when previously detected
    trust topology conflict is automatically solved.

    No separate errno is assigned as this error should
    not be visible outside the dcerpc.py code.
    """


def assess_dcerpc_error(error):
    """
    Takes error returned by Samba bindings and converts it into
    an IPA error class.
    """
    if isinstance(error, RuntimeError):
        error_tuple = error.args
    else:
        error_tuple = error
    if len(error_tuple) != 2:
        raise RuntimeError("Unable to parse error: {err!r}".format(err=error))

    num, message = error_tuple
    if num and num in dcerpc_error_codes:
        return dcerpc_error_codes[num]
    if message and message in dcerpc_error_messages:
        return dcerpc_error_messages[message]
    reason = _('CIFS server communication error: code "%(num)s", '
               'message "%(message)s" (both may be "None")') % \
        dict(num=num, message=message)
    return errors.RemoteRetrieveError(reason=reason)


class ExtendedDNControl(LDAPControl):
    def __init__(self):
        LDAPControl.__init__(
            self,
            controlType="1.2.840.113556.1.4.529",
            criticality=False,
            encodedControlValue=b'0\x03\x02\x01\x01'
        )


class DomainValidator:
    ATTR_FLATNAME = 'ipantflatname'
    ATTR_SID = 'ipantsecurityidentifier'
    ATTR_TRUSTED_SID = 'ipanttrusteddomainsid'
    ATTR_TRUST_PARTNER = 'ipanttrustpartner'
    ATTR_TRUST_AUTHOUT = 'ipanttrustauthoutgoing'

    def __init__(self, api):
        self.api = api
        self.ldap = self.api.Backend.ldap2
        self.domain = None
        self.flatname = None
        self.dn = None
        self.sid = None
        self._domains = None
        self._info = dict()
        self._creds = None
        self._admin_creds = None
        self._parm = None

    def is_configured(self):
        cn_trust_local = DN(('cn', self.api.env.domain),
                            self.api.env.container_cifsdomains,
                            self.api.env.basedn)
        try:
            entry_attrs = self.ldap.get_entry(cn_trust_local,
                                              [self.ATTR_FLATNAME,
                                               self.ATTR_SID])
            self.flatname = entry_attrs[self.ATTR_FLATNAME][0]
            self.sid = entry_attrs[self.ATTR_SID][0]
            self.dn = entry_attrs.dn
            self.domain = self.api.env.domain
        except errors.NotFound:
            return False
        return True

    def get_trusted_domains(self):
        """
        Returns case-insensitive dict of trusted domain tuples
        (flatname, sid, trust_auth_outgoing), keyed by domain name.
        """
        cn_trust = DN(('cn', 'ad'), self.api.env.container_trusts,
                      self.api.env.basedn)

        try:
            search_kw = {'objectClass': 'ipaNTTrustedDomain'}
            filter = self.ldap.make_filter(search_kw,
                                           rules=self.ldap.MATCH_ALL)
            entries, _truncated = self.ldap.find_entries(
                filter=filter,
                base_dn=cn_trust,
                attrs_list=[self.ATTR_TRUSTED_SID,
                            self.ATTR_FLATNAME,
                            self.ATTR_TRUST_PARTNER]
                )

            # We need to use case-insensitive dictionary since we use
            # domain names as keys and those are generally case-insensitive
            result = ipautil.CIDict()

            for e in entries:
                try:
                    t_partner = e.single_value.get(self.ATTR_TRUST_PARTNER)
                    fname_norm = e.single_value.get(self.ATTR_FLATNAME).lower()
                    trusted_sid = e.single_value.get(self.ATTR_TRUSTED_SID)
                except KeyError as exc:
                    # Some piece of trusted domain info in LDAP is missing
                    # Skip the domain, but leave log entry for investigation
                    logger.warning("Trusted domain '%s' entry misses an "
                                   "attribute: %s", e.dn, exc)
                    continue

                result[t_partner] = (fname_norm,
                                     security.dom_sid(trusted_sid))
            return result
        except errors.NotFound as exc:
            return []

    def set_trusted_domains(self):
        # At this point we have SID_NT_AUTHORITY family SID and really need to
        # check it against prefixes of domain SIDs we trust to
        if not self._domains:
            self._domains = self.get_trusted_domains()
        if len(self._domains) == 0:
            # Our domain is configured but no trusted domains are configured
            # This means we can't check the correctness of a trusted
            # domain SIDs
            raise errors.ValidationError(name='sid',
                                         error=_('no trusted domain '
                                                 'is configured'))

    def get_domain_by_sid(self, sid, exact_match=False):
        if not self.domain:
            # our domain is not configured or self.is_configured() never run
            # reject SIDs as we can't check correctness of them
            raise errors.ValidationError(name='sid',
                                         error=_('domain is not configured'))

        # Parse sid string to see if it is really in a SID format
        try:
            test_sid = security.dom_sid(sid)
        except TypeError:
            raise errors.ValidationError(name='sid',
                                         error=_('SID is not valid'))

        # At this point we have SID_NT_AUTHORITY family SID and really need to
        # check it against prefixes of domain SIDs we trust to
        self.set_trusted_domains()

        # We have non-zero list of trusted domains and have to go through
        # them one by one and check their sids as prefixes / exact match
        # depending on the value of exact_match flag
        if exact_match:
            # check exact match of sids
            for domain in self._domains:
                if sid == str(self._domains[domain][1]):
                    return domain

            raise errors.NotFound(reason=_("SID does not match exactly"
                                           "with any trusted domain's SID"))
        else:
            # check as prefixes
            test_sid_subauths = test_sid.sub_auths
            for domain in self._domains:
                domsid = self._domains[domain][1]
                sub_auths = domsid.sub_auths
                num_auths = min(test_sid.num_auths, domsid.num_auths)
                if test_sid_subauths[:num_auths] == sub_auths[:num_auths]:
                    return domain
            raise errors.NotFound(reason=_('SID does not match any '
                                           'trusted domain'))

    def is_trusted_sid_valid(self, sid):
        try:
            self.get_domain_by_sid(sid)
        except (errors.ValidationError, errors.NotFound):
            return False
        else:
            return True

    def is_trusted_domain_sid_valid(self, sid):
        try:
            self.get_domain_by_sid(sid, exact_match=True)
        except (errors.ValidationError, errors.NotFound):
            return False
        else:
            return True

    def get_sid_from_domain_name(self, name):
        """Returns binary representation of SID for the trusted domain name
           or None if name is not in the list of trusted domains."""

        domains = self.get_trusted_domains()
        if name in domains:
            return domains[name][1]
        else:
            return None

    def get_trusted_domain_objects(self, domain=None, flatname=None, filter="",
                                   attrs=None, scope=_ldap.SCOPE_SUBTREE,
                                   basedn=None):
        """
        Search for LDAP objects in a trusted domain specified either by
        `domain' or `flatname'. The actual LDAP search is specified by
        `filter', `attrs', `scope' and `basedn'. When `basedn' is empty,
        database root DN is used.
        """
        assert domain is not None or flatname is not None
        """Returns SID for the trusted domain object (user or group only)"""
        if not self.domain:
            # our domain is not configured or self.is_configured() never run
            raise errors.ValidationError(name=_('Trust setup'),
                                         error=_('Our domain is '
                                                 'not configured'))
        if not self._domains:
            self._domains = self.get_trusted_domains()
        if len(self._domains) == 0:
            # Our domain is configured but no trusted domains are configured
            raise errors.ValidationError(name=_('Trust setup'),
                                         error=_('No trusted domain is '
                                                 'not configured'))

        entries = None
        if domain is not None:
            if domain not in self._domains:
                raise errors.ValidationError(name=_('trusted domain object'),
                                             error=_('domain is not trusted'))
            # Now we have a name to check against our list of trusted domains
            entries = self.search_in_dc(domain, filter, attrs, scope, basedn)
        elif flatname is not None:
            # Flatname was specified, traverse through the list of trusted
            # domains first to find the proper one
            found_flatname = False
            for domain in self._domains:
                if self._domains[domain][0] == flatname:
                    found_flatname = True
                    entries = self.search_in_dc(domain, filter,
                                                attrs, scope, basedn)
                    if entries:
                        break
            if not found_flatname:
                raise errors.ValidationError(name=_('trusted domain object'),
                                             error=_('no trusted domain '
                                                     'matched the specified '
                                                     'flat name'))
        if not entries:
            raise errors.NotFound(reason=_('trusted domain object not found'))

        return entries

    def get_trusted_domain_object_sid(self, object_name,
                                      fallback_to_ldap=True):
        result = pysss_nss_idmap.getsidbyname(object_name)
        if object_name in result and \
           (pysss_nss_idmap.SID_KEY in result[object_name]):
            object_sid = result[object_name][pysss_nss_idmap.SID_KEY]
            return object_sid

        # If fallback to AD DC LDAP is not allowed, bail out
        if not fallback_to_ldap:
            raise errors.ValidationError(name=_('trusted domain object'),
                                         error=_('SSSD was unable to resolve '
                                                 'the object to a valid SID'))

        # Else, we are going to contact AD DC LDAP
        components = normalize_name(object_name)
        if not ('domain' in components or 'flatname' in components):
            # No domain or realm specified, ambiguous search
            raise errors.ValidationError(name=_('trusted domain object'),
                                         error=_('Ambiguous search, user '
                                                 'domain was not specified'))

        attrs = ['objectSid']
        filter = '(&(sAMAccountName=%(name)s)' \
                 '(|(objectClass=user)(objectClass=group)))' \
                 % dict(name=components['name'])
        scope = _ldap.SCOPE_SUBTREE
        entries = self.get_trusted_domain_objects(components.get('domain'),
                                                  components.get('flatname'),
                                                  filter, attrs, scope)

        if len(entries) > 1:
            # Treat non-unique entries as invalid
            raise errors.ValidationError(name=_('trusted domain object'),
                                         error=_('Trusted domain did not '
                                                 'return a unique object'))
        sid = self.__sid_to_str(entries[0]['objectSid'][0])
        try:
            test_sid = security.dom_sid(sid)
            return unicode(test_sid)
        except TypeError:
            raise errors.ValidationError(name=_('trusted domain object'),
                                         error=_('Trusted domain did not '
                                                 'return a valid SID for '
                                                 'the object'))

    def get_trusted_domain_object_type(self, name_or_sid):
        """
        Return the type of the object corresponding to the given name in
        the trusted domain, which is either 'user', 'group' or 'both'.
        The 'both' types is used for users with magic private groups.
        """

        object_type = None

        if is_sid_valid(name_or_sid):
            result = pysss_nss_idmap.getnamebysid(name_or_sid)
        else:
            result = pysss_nss_idmap.getsidbyname(name_or_sid)

        if name_or_sid in result:
            object_type = result[name_or_sid].get(pysss_nss_idmap.TYPE_KEY)

        # Do the translation to hide pysss_nss_idmap constants
        # from higher-level code
        return pysss_type_key_translation_dict.get(object_type)

    def get_trusted_domain_object_from_sid(self, sid):
        logger.debug("Converting SID to object name: %s", sid)

        # Check if the given SID is valid
        if not self.is_trusted_sid_valid(sid):
            raise errors.ValidationError(name='sid', error='SID is not valid')

        # Use pysss_nss_idmap to obtain the name
        result = pysss_nss_idmap.getnamebysid(sid).get(sid)

        valid_types = (pysss_nss_idmap.ID_USER,
                       pysss_nss_idmap.ID_GROUP,
                       pysss_nss_idmap.ID_BOTH)

        if result:
            if result.get(pysss_nss_idmap.TYPE_KEY) in valid_types:
                return result.get(pysss_nss_idmap.NAME_KEY)

        # If unsuccessful, search AD DC LDAP
        logger.debug("Searching AD DC LDAP")

        escaped_sid = escape_filter_chars(
            security.dom_sid(sid).__ndr_pack__(),
            2  # 2 means every character needs to be escaped
        )

        attrs = ['sAMAccountName']
        filter = (r'(&(objectSid=%(sid)s)'
                  '(|(objectClass=user)(objectClass=group)))'
                  % dict(sid=escaped_sid))  # sid in binary
        domain = self.get_domain_by_sid(sid)

        entries = self.get_trusted_domain_objects(domain=domain,
                                                  filter=filter,
                                                  attrs=attrs)

        if len(entries) > 1:
            # Treat non-unique entries as invalid
            raise errors.ValidationError(name=_('trusted domain object'),
                                         error=_('Trusted domain did not '
                                                 'return a unique object'))

        object_name = (
            "%s@%s" % (entries[0].single_value['sAMAccountName'].lower(),
                       domain.lower())
            )

        return unicode(object_name)

    def __get_trusted_domain_user_and_groups(self, object_name):
        """
        Returns a tuple with user SID and a list of SIDs of all groups he is
        a member of.

        LIMITATIONS:
            - only Trusted Admins group members can use this function as it
              uses secret for IPA-Trusted domain link
            - List of group SIDs does not contain group memberships outside
              of the trusted domain
        """
        components = normalize_name(object_name)
        domain = components.get('domain')
        flatname = components.get('flatname')
        name = components.get('name')

        is_valid_sid = is_sid_valid(object_name)
        if is_valid_sid:
            # Find a trusted domain for the SID
            domain = self.get_domain_by_sid(object_name)
            # Now search a trusted domain for a user with this SID
            attrs = ['cn']
            filter = '(&(objectClass=user)(objectSid=%(sid)s))' \
                % dict(sid=object_name)
            try:
                entries = self.get_trusted_domain_objects(domain=domain,
                                                          filter=filter,
                                                          attrs=attrs,
                                                          scope=_ldap.SCOPE_SUBTREE)
            except errors.NotFound:
                raise errors.NotFound(reason=_('trusted domain user not found'))
            user_dn = entries[0].dn
        elif domain or flatname:
            attrs = ['cn']
            filter = '(&(sAMAccountName=%(name)s)(objectClass=user))' \
                     % dict(name=name)
            try:
                entries = self.get_trusted_domain_objects(domain,
                                                          flatname, filter, attrs,
                                                          _ldap.SCOPE_SUBTREE)
            except errors.NotFound:
                raise errors.NotFound(reason=_('trusted domain user not found'))
            user_dn = entries[0].dn
        else:
            # No domain or realm specified, ambiguous search
            raise errors.ValidationError(name=_('trusted domain object'),
                                         error=_('Ambiguous search, '
                                                 'user domain was not specified'))

        # Get SIDs of user object and it's groups
        # tokenGroups attribute must be read with a scope BASE for a known user
        # distinguished name to avoid search error
        attrs = ['objectSID', 'tokenGroups']
        filter = "(objectClass=user)"
        entries = self.get_trusted_domain_objects(domain,
                                                  flatname, filter, attrs,
                                                  _ldap.SCOPE_BASE, user_dn)
        object_sid = self.__sid_to_str(entries[0]['objectSid'][0])
        group_sids = [self.__sid_to_str(sid)
                      for sid in entries[0]['tokenGroups']]
        return (object_sid, group_sids)

    def get_trusted_domain_user_and_groups(self, object_name):
        """
        Returns a tuple with user SID and a list of SIDs of all groups he is
        a member of.

        First attempts to perform SID lookup via SSSD and in case of failure
        resorts back to checking trusted domain's AD DC LDAP directly.

        LIMITATIONS:
            - only Trusted Admins group members can use this function as it
              uses secret for IPA-Trusted domain link if SSSD lookup failed
            - List of group SIDs does not contain group memberships outside
              of the trusted domain
        """
        group_sids = None
        group_list = None
        object_sid = None
        is_valid_sid = is_sid_valid(object_name)
        if is_valid_sid:
            object_sid = object_name
            result = pysss_nss_idmap.getnamebysid(object_name)
            if object_name in result and \
               (pysss_nss_idmap.NAME_KEY in result[object_name]):
                group_list = pysss.getgrouplist(
                                 result[object_name][pysss_nss_idmap.NAME_KEY])
        else:
            result = pysss_nss_idmap.getsidbyname(object_name)
            if object_name in result and \
               (pysss_nss_idmap.SID_KEY in result[object_name]):
                object_sid = result[object_name][pysss_nss_idmap.SID_KEY]
                group_list = pysss.getgrouplist(object_name)

        if not group_list:
            return self.__get_trusted_domain_user_and_groups(object_name)

        group_sids = pysss_nss_idmap.getsidbyname(group_list)
        return (
                object_sid,
                [el[1][pysss_nss_idmap.SID_KEY] for el in group_sids.items()]
               )

    def __sid_to_str(self, sid):
        """
        Converts binary SID to string representation
        Returns unicode string
        """
        sid_rev_num = ord(sid[0])
        number_sub_id = ord(sid[1])
        ia = struct.unpack('!Q', '\x00\x00'+sid[2:8])[0]
        subs = [
            struct.unpack('<I', sid[8+4*i:12+4*i])[0]
            for i in range(number_sub_id)
        ]
        return u'S-%d-%d-%s' % (sid_rev_num, ia,
                                '-'.join([str(s) for s in subs]),)

    def kinit_as_administrator(self, domain):
        """
        Initializes ccache with http service credentials.

        Applies session code defaults for ccache directory and naming prefix.
        Session code uses kinit_+<pid>, we use
        kinit_+<TD>+<domain netbios name> so there is no clash.

        Returns tuple (ccache path, principal) where (None, None) signifes an
        error on ccache initialization
        """

        if self._admin_creds is None:
            return (None, None)

        domain_suffix = domain.replace('.', '-')

        ccache_name = "kinit_TDA%s" % (domain_suffix)
        ccache_path = os.path.join(paths.IPA_CCACHES, ccache_name)

        (principal, password) = self._admin_creds.split('%', 1)

        # Destroy the contents of the ccache
        logger.debug('Destroying the contents of the separate ccache')

        ipautil.run(
            [paths.KDESTROY, '-A', '-c', ccache_path],
            env={'KRB5CCNAME': ccache_path},
            raiseonerr=False)

        # Destroy the contents of the ccache
        logger.debug('Running kinit with credentials of AD administrator')

        result = ipautil.run(
            [paths.KINIT, principal],
            env={'KRB5CCNAME': ccache_path},
            stdin=password,
            raiseonerr=False)

        if result.returncode == 0:
            return (ccache_path, principal)
        else:
            return (None, None)

    def search_in_dc(self, domain, filter, attrs, scope, basedn=None,
                     quiet=False):
        """
        Perform LDAP search in a trusted domain `domain' Domain Controller.
        Returns resulting entries or None.
        """

        entries = None

        info = self.__retrieve_trusted_domain_gc_list(domain)

        if not info:
            raise errors.ValidationError(
                name=_('Trust setup'),
                error=_('Cannot retrieve trusted domain GC list'))

        for (host, port) in info['gc']:
            entries = self.__search_in_dc(info, host, port, filter, attrs,
                                          scope, basedn=basedn,
                                          quiet=quiet)
            if entries:
                break

        return entries

    def __search_in_dc(self, info, host, port, filter, attrs, scope,
                       basedn=None, quiet=False):
        """
        Actual search in AD LDAP server, using SASL GSSAPI authentication
        Returns LDAP result or None.
        """

        ccache_name = None

        if self._admin_creds:
            (ccache_name,
             _principal) = self.kinit_as_administrator(info['dns_domain'])

        if ccache_name:
            with ipautil.private_ccache(path=ccache_name):
                entries = None

                try:
                    # AD does not support SASL + TLS at the same time
                    # https://msdn.microsoft.com/en-us/library/cc223500.aspx
                    conn = ipaldap.LDAPClient.from_hostname_plain(
                        host,
                        no_schema=True,
                        decode_attrs=False
                    )
                    conn.gssapi_bind()

                    if basedn is None:
                        # Use domain root base DN
                        basedn = ipautil.realm_to_suffix(info['dns_domain'])

                    entries = conn.get_entries(basedn, scope, filter, attrs)
                except Exception as e:
                    msg = "Search on AD DC {host}:{port} failed with: {err}"\
                          .format(host=host, port=str(port), err=str(e))
                    if quiet:
                        logger.debug('%s', msg)
                    else:
                        logger.warning('%s', msg)

                return entries
        return None

    def __retrieve_trusted_domain_gc_list(self, domain):
        """
        Retrieves domain information and preferred GC list
        Returns dictionary with following keys
             name       -- NetBIOS name of the trusted domain
             dns_domain -- DNS name of the trusted domain
             gc         -- array of tuples (server, port) for Global Catalog
        """
        if domain in self._info:
            return self._info[domain]

        if not self._creds:
            self._parm = param.LoadParm()
            self._parm.load(
                os.path.join(paths.USR_SHARE_IPA_DIR, "smb.conf.empty"))
            self._parm.set('netbios name', self.flatname)
            self._creds = credentials.Credentials()
            self._creds.set_kerberos_state(credentials.MUST_USE_KERBEROS)
            self._creds.guess(self._parm)
            self._creds.set_workstation(self.flatname)

        netrc = net.Net(creds=self._creds, lp=self._parm)
        finddc_error = None
        result = None
        flags = nbt.NBT_SERVER_LDAP | nbt.NBT_SERVER_GC | nbt.NBT_SERVER_CLOSEST
        try:
            result = netrc.finddc(domain=domain, flags=flags)
        except RuntimeError as e:
            try:
                # If search of closest GC failed, attempt to find any one
                flags = nbt.NBT_SERVER_LDAP | nbt.NBT_SERVER_GC
                result = netrc.finddc(domain=domain, flags=flags)
            except RuntimeError as e:
                finddc_error = e

        if not self._domains:
            self._domains = self.get_trusted_domains()

        info = dict()
        servers = []

        if result:
            info['name'] = unicode(result.domain_name)
            info['dns_domain'] = unicode(result.dns_domain)
            servers = [(unicode(result.pdc_dns_name), 3268)]
        else:
            info['name'] = self._domains[domain]
            info['dns_domain'] = domain
            # Retrieve GC servers list
            gc_name = '_gc._tcp.%s.' % info['dns_domain']

            try:
                answers = query_srv(gc_name)
            except DNSException as e:
                answers = []

            for answer in answers:
                server = str(answer.target).rstrip(".")
                servers.append((server, answer.port))

        info['gc'] = servers

        # Both methods should not fail at the same time
        if finddc_error and len(info['gc']) == 0:
            raise assess_dcerpc_error(finddc_error)

        self._info[domain] = info
        return info


def string_to_array(what):
    if six.PY3 and isinstance(what, bytes):
        return [v for v in what]
    return [ord(v) for v in what]


class TrustDomainInstance:

    def __init__(self, hostname, creds=None):
        self.parm = param.LoadParm()
        self.parm.load(os.path.join(paths.USR_SHARE_IPA_DIR, "smb.conf.empty"))
        if len(hostname) > 0:
            self.parm.set('netbios name', hostname)
        self.creds = creds
        self.hostname = hostname
        self.info = {}
        self._pipe = None
        self._policy_handle = None
        self.read_only = False
        self.ftinfo_records = None
        self.validation_attempts = 0

    def __gen_lsa_connection(self, binding):
        if self.creds is None:
            raise errors.RequirementError(name=_('CIFS credentials object'))
        try:
            result = lsa.lsarpc(binding, self.parm, self.creds)
            return result
        except RuntimeError as e:
            raise assess_dcerpc_error(e)

    def init_lsa_pipe(self, remote_host):
        """
        Try to initialize connection to the LSA pipe at remote host.
        This method tries consequently all possible transport options
        and selects one that works. See __gen_lsa_bindings() for details.

        The actual result may depend on details of existing credentials.
        For example, using signing causes NO_SESSION_KEY with Win2K8 and
        using kerberos against Samba with signing does not work.
        """
        # short-cut: if LSA pipe is initialized, skip completely
        if self._pipe:
            return

        attempts = 0
        session_attempts = 0
        bindings = self.__gen_lsa_bindings(remote_host)
        for binding in bindings:
            try:
                self._pipe = self.__gen_lsa_connection(binding)
                if self._pipe and self._pipe.session_key:
                    break
            except errors.ACIError:
                attempts = attempts + 1
            except RuntimeError:
                # When session key is not available, we just skip this binding
                session_attempts = session_attempts + 1

        if self._pipe is None and \
           (attempts + session_attempts) == len(bindings):
            raise errors.ACIError(
                      info=_('CIFS server %(host)s denied your credentials')
                      % dict(host=remote_host))

        if self._pipe is None:
            raise errors.RemoteRetrieveError(
                    reason=_('Cannot establish LSA connection to %(host)s. '
                             'Is CIFS server running?') % dict(host=remote_host))
        self.binding = binding
        self.session_key = self._pipe.session_key

    def __gen_lsa_bindings(self, remote_host):
        """
        There are multiple transports to issue LSA calls. However, depending on
        a system in use they may be blocked by local operating system policies.
        Generate all we can use. init_lsa_pipe() will try them one by one until
        there is one working.

        We try NCACN_NP before NCACN_IP_TCP and use SMB2 before SMB1.
        """
        transports = (u'ncacn_np', u'ncacn_ip_tcp')
        options = (u'smb2,print', u'print')
        return [u'%s:%s[%s]' % (t, remote_host, o)
                for t in transports for o in options]

    def retrieve_anonymously(self, remote_host,
                             discover_srv=False, search_pdc=False):
        """
        When retrieving DC information anonymously, we can't get SID of the domain
        """
        netrc = net.Net(creds=self.creds, lp=self.parm)
        flags = nbt.NBT_SERVER_LDAP | nbt.NBT_SERVER_DS | nbt.NBT_SERVER_WRITABLE
        if search_pdc:
            flags = flags | nbt.NBT_SERVER_PDC
        try:
            if discover_srv:
                result = netrc.finddc(domain=remote_host, flags=flags)
            else:
                result = netrc.finddc(address=remote_host, flags=flags)
        except RuntimeError as e:
            raise assess_dcerpc_error(e)

        if not result:
            return False
        self.info['name'] = unicode(result.domain_name)
        self.info['dns_domain'] = unicode(result.dns_domain)
        self.info['dns_forest'] = unicode(result.forest)
        self.info['guid'] = unicode(result.domain_uuid)
        self.info['dc'] = unicode(result.pdc_dns_name)
        self.info['is_pdc'] = (result.server_type & nbt.NBT_SERVER_PDC) != 0

        # Netlogon response doesn't contain SID of the domain.
        # We need to do rootDSE search with LDAP_SERVER_EXTENDED_DN_OID
        # control to reveal the SID
        ldap_uri = 'ldap://%s' % (result.pdc_dns_name)
        conn = ldap_initialize(ldap_uri)
        conn.set_option(_ldap.OPT_SERVER_CONTROLS, [ExtendedDNControl()])
        search_result = None
        try:
            _objtype, res = conn.search_s('', _ldap.SCOPE_BASE)[0]
            for o in res.keys():
                if isinstance(res[o], list):
                    t = res[o]
                    for z, v in enumerate(t):
                        if isinstance(v, bytes):
                            t[z] = v.decode('utf-8')
                elif isinstance(res[o], bytes):
                    res[o] = res[o].decode('utf-8')
            search_result = res['defaultNamingContext'][0]
            self.info['dns_hostname'] = res['dnsHostName'][0]
        except _ldap.LDAPError as e:
            logger.error(
                "LDAP error when connecting to %s: %s",
                unicode(result.pdc_name), str(e))
        except KeyError as e:
            logger.error("KeyError: %s, LDAP entry from %s "
                         "returned malformed. Your DNS might be "
                         "misconfigured.",
                         unicode(e),
                         unicode(result.pdc_name))

        if search_result:
            self.info['sid'] = self.parse_naming_context(search_result)
        return True

    def parse_naming_context(self, context):
        naming_ref = re.compile('.*<SID=(S-.*)>.*')
        return unicode(naming_ref.match(context).group(1))

    def retrieve(self, remote_host):
        self.init_lsa_pipe(remote_host)

        objectAttribute = lsa.ObjectAttribute()
        objectAttribute.sec_qos = lsa.QosInfo()
        try:
            self._policy_handle = \
                self._pipe.OpenPolicy2(u"", objectAttribute,
                                       security.SEC_FLAG_MAXIMUM_ALLOWED)
            result = self._pipe.QueryInfoPolicy2(self._policy_handle,
                                                 lsa.LSA_POLICY_INFO_DNS)
        except RuntimeError as e:
            raise assess_dcerpc_error(e)

        self.info['name'] = unicode(result.name.string)
        self.info['dns_domain'] = unicode(result.dns_domain.string)
        self.info['dns_forest'] = unicode(result.dns_forest.string)
        self.info['guid'] = unicode(result.domain_guid)
        self.info['sid'] = unicode(result.sid)
        self.info['dc'] = remote_host

        try:
            result = self._pipe.QueryInfoPolicy2(self._policy_handle,
                                                 lsa.LSA_POLICY_INFO_ROLE)
        except RuntimeError as e:
            raise assess_dcerpc_error(e)

        self.info['is_pdc'] = (result.role == lsa.LSA_ROLE_PRIMARY)

    def generate_auth(self, trustdom_secret):
        password_blob = string_to_array(trustdom_secret.encode('utf-16-le'))

        clear_value = drsblobs.AuthInfoClear()
        clear_value.size = len(password_blob)
        clear_value.password = password_blob

        clear_authinfo = drsblobs.AuthenticationInformation()
        clear_authinfo.LastUpdateTime = samba.unix2nttime(int(time.time()))
        clear_authinfo.AuthType = lsa.TRUST_AUTH_TYPE_CLEAR
        clear_authinfo.AuthInfo = clear_value

        authinfo_array = drsblobs.AuthenticationInformationArray()
        authinfo_array.count = 1
        authinfo_array.array = [clear_authinfo]

        outgoing = drsblobs.trustAuthInOutBlob()
        outgoing.count = 1
        outgoing.current = authinfo_array

        confounder = [3]*512
        for i in range(512):
            confounder[i] = random.randint(0, 255)

        trustpass = drsblobs.trustDomainPasswords()
        trustpass.confounder = confounder

        trustpass.outgoing = outgoing
        trustpass.incoming = outgoing

        trustpass_blob = ndr_pack(trustpass)

        encrypted_trustpass = arcfour_encrypt(self._pipe.session_key,
                                              trustpass_blob)

        auth_blob = lsa.DATA_BUF2()
        auth_blob.size = len(encrypted_trustpass)
        auth_blob.data = string_to_array(encrypted_trustpass)

        auth_info = lsa.TrustDomainInfoAuthInfoInternal()
        auth_info.auth_blob = auth_blob
        self.auth_info = auth_info

    def generate_ftinfo(self, another_domain):
        """
        Generates TrustDomainInfoFullInfo2Internal structure
        This structure allows to pass information about all domains associated
        with the another domain's realm.

        Only top level name and top level name exclusions are handled here.
        """
        if not another_domain.ftinfo_records:
            return None

        ftinfo_records = []
        info = lsa.ForestTrustInformation()

        for rec in another_domain.ftinfo_records:
            record = lsa.ForestTrustRecord()
            record.flags = 0
            record.time = rec['rec_time']
            record.type = rec['rec_type']
            record.forest_trust_data.string = rec['rec_name']
            ftinfo_records.append(record)

        info.count = len(ftinfo_records)
        info.entries = ftinfo_records
        return info

    def clear_ftinfo_conflict(self, another_domain, cinfo):
        """
        Attempt to clean up the forest trust collisions

        :param self: the forest we establish trust to
        :param another_domain: a forest that establishes trust to 'self'
        :param cinfo: lsa_ForestTrustCollisionInfo structure that contain
                      set of of lsa_ForestTrustCollisionRecord structures
        :raises: TrustTopologyConflictSolved, TrustTopologyConflictError

        This code tries to perform intelligent job of going
        over individual collisions and making exclusion entries
        for affected IPA namespaces.

        There are three possible conflict configurations:
          - conflict of DNS namespace (TLN conflict, LSA_TLN_DISABLED_CONFLICT)
          - conflict of SID namespace (LSA_SID_DISABLED_CONFLICT)
          - conflict of NetBIOS namespace (LSA_NB_DISABLED_CONFLICT)

        we only can handle TLN conflicts because (a) excluding SID namespace
        is not possible and (b) excluding NetBIOS namespace not possible.
        These two types of conflicts should result in trust-add CLI error

        These conflicts can come from external source (another forest) or
        from internal source (another domain in the same forest). We only
        can fix the problems with another forest.

        To resolve TLN conflict we need to do following:
          1. Retrieve forest trust information for the forest we conflict on
          2. Add an exclusion entry for IPA DNS namespace to it
          3. Set forest trust information for the forest we conflict on
          4. Re-try establishing trust to the original forest

        This all can only be done under privileges of Active Directory admin
        that can change forest trusts. If we cannot have those privileges,
        the work has to be done manually in the Windows UI for
        'Active Directory Domains and Trusts' by the administrator of the
        original forest.
        """

        def domain_name_from_ftinfo(ftinfo):
            """
            Returns a domain name string from a ForestTrustRecord

            :param ftinfo: LSA ForestTrustRecord to parse
            """
            if ftinfo.type == lsa.LSA_FOREST_TRUST_DOMAIN_INFO:
                return ftinfo.forest_trust_data.dns_domain_name.string
            elif ftinfo.type == lsa.LSA_FOREST_TRUST_TOP_LEVEL_NAME:
                return ftinfo.forest_trust_data.string
            elif ftinfo.type == lsa.LSA_FOREST_TRUST_TOP_LEVEL_NAME_EX:
                # We should ignore TLN exclusion record because it
                # is already an exclusion so we aren't going to
                # change anything here
                return None
            else:
                # Ignore binary blobs we don't know about
                return None

        # List of entries for unsolved conflicts
        result = []

        trust_timestamp = long(time.time()*1e7+116444736000000000)

        # Collision information contains entries for specific trusted domains
        # we collide with. Look into TLN collisions and add a TLN exclusion
        # entry to the specific domain trust.
        logger.error("Attempt to solve forest trust topology conflicts")
        for rec in cinfo.entries:
            if rec.type == lsa.LSA_FOREST_TRUST_COLLISION_TDO:
                dominfo = self._pipe.lsaRQueryForestTrustInformation(
                                 self._policy_handle,
                                 rec.name,
                                 lsa.LSA_FOREST_TRUST_DOMAIN_INFO)

                # Oops, we were unable to retrieve trust topology for this
                # trusted domain (forest).
                if not dominfo:
                    result.append(rec)
                    logger.error("Unable to resolve conflict for "
                                 "DNS domain %s in the forest %s "
                                 "for domain trust %s. Trust cannot "
                                 "be established unless this conflict "
                                 "is fixed manually.",
                                 another_domain.info['dns_domain'],
                                 self.info['dns_domain'],
                                 rec.name.string)
                    continue

                # Copy over the entries, extend with TLN exclusion
                entries = []
                is_our_record = False
                for e in dominfo.entries:
                    e1 = lsa.ForestTrustRecord()
                    e1.type = e.type
                    e1.flags = e.flags
                    e1.time = e.time
                    e1.forest_trust_data = e.forest_trust_data

                    # We either have a domain struct, a TLN name,
                    # or a TLN exclusion name in the list.
                    # The rest we should skip, those are binary blobs
                    dns_domain_name = domain_name_from_ftinfo(e)

                    # Search for a match in the topology of another domain
                    # if there is a match, we have to convert a record
                    # into a TLN exclusion to allow its routing to the
                    # another domain
                    for r in another_domain.ftinfo_records:
                        # r['rec_name'] cannot be None, thus we can ignore
                        # the case when dns_domain_name is None
                        if r['rec_name'] == dns_domain_name:
                            is_our_record = True

                            # Convert e1 into an exclusion record
                            e1.type = lsa.LSA_FOREST_TRUST_TOP_LEVEL_NAME_EX
                            e1.flags = 0
                            e1.time = trust_timestamp
                            e1.forest_trust_data.string = dns_domain_name
                            break
                    entries.append(e1)

                # If no candidate for the exclusion entry was found
                # make sure it is the other domain itself, this covers
                # a most common case
                if not is_our_record:
                    # Create TLN exclusion record for the top level domain
                    record = lsa.ForestTrustRecord()
                    record.type = lsa.LSA_FOREST_TRUST_TOP_LEVEL_NAME_EX
                    record.flags = 0
                    record.time = trust_timestamp
                    record.forest_trust_data.string = \
                        another_domain.info['dns_domain']
                    entries.append(record)

                fti = lsa.ForestTrustInformation()
                fti.count = len(entries)
                fti.entries = entries

                # Update the forest trust information now
                ldname = lsa.StringLarge()
                ldname.string = rec.name.string
                cninfo = None
                try:
                    cninfo = self._pipe.lsaRSetForestTrustInformation(
                        self._policy_handle,
                        ldname,
                        lsa.LSA_FOREST_TRUST_DOMAIN_INFO,
                        fti, 0)
                except samba.NTSTATUSError as error:
                    # Handle NT_STATUS_INVALID_PARAMETER separately
                    if ntstatus.NT_STATUS_INVALID_PARAMETER == error.args[0]:
                        result.append(rec)
                        logger.error("Unable to resolve conflict for "
                                     "DNS domain %s in the forest %s "
                                     "for in-forest domain %s. Trust cannot "
                                     "be established unless this conflict "
                                     "is fixed manually.",
                                     another_domain.info['dns_domain'],
                                     self.info['dns_domain'],
                                     rec.name.string)
                    else:
                        raise assess_dcerpc_error(error)


                if cninfo:
                    result.append(rec)
                    logger.error("When defining exception for DNS "
                                 "domain %s in forest %s for "
                                 "trusted forest %s, "
                                 "got collision info back:\n%s",
                                 another_domain.info['dns_domain'],
                                 self.info['dns_domain'],
                                 rec.name.string,
                                 ndr_print(cninfo))
            else:
                result.append(rec)
                logger.error("Unable to resolve conflict for "
                             "DNS domain %s in the forest %s "
                             "for in-forest domain %s. Trust cannot "
                             "be established unless this conflict "
                             "is fixed manually.",
                             another_domain.info['dns_domain'],
                             self.info['dns_domain'],
                             rec.name.string)

        if len(result) == 0:
            logger.error("Successfully solved all conflicts")
            raise TrustTopologyConflictSolved()

        # Otherwise, raise TrustTopologyConflictError() exception
        domains = [x.name.string for x in result]
        raise errors.TrustTopologyConflictError(
            forest=self.info['dns_domain'],
            conflict=another_domain.info['dns_domain'],
            domains=domains)



    def update_ftinfo(self, another_domain):
        """
        Updates forest trust information in this forest corresponding
        to the another domain's information.
        """
        if another_domain.ftinfo_records:
            ftinfo = self.generate_ftinfo(another_domain)
            # Set forest trust information -- we do it only against AD DC as
            # smbd already has the information about itself
            ldname = lsa.StringLarge()
            ldname.string = another_domain.info['dns_domain']
            ftlevel = lsa.LSA_FOREST_TRUST_DOMAIN_INFO
            # RSetForestTrustInformation returns collision information
            # for trust topology
            cinfo = self._pipe.lsaRSetForestTrustInformation(
                        self._policy_handle,
                        ldname,
                        ftlevel,
                        ftinfo, 0)
            if cinfo:
                logger.error("When setting forest trust information, "
                             "got collision info back:\n%s",
                             ndr_print(cinfo))
                self.clear_ftinfo_conflict(another_domain, cinfo)

    def establish_trust(self, another_domain, trustdom_secret,
                        trust_type='bidirectional', trust_external=False):
        """
        Establishes trust between our and another domain
        Input: another_domain -- instance of TrustDomainInstance,
                                 initialized with #retrieve call
               trustdom_secret -- shared secred used for the trust
        """
        if self.info['name'] == another_domain.info['name']:
            # Check that NetBIOS names do not clash
            raise errors.ValidationError(name=u'AD Trust Setup',
                                         error=_('the IPA server and the '
                                                 'remote domain cannot share '
                                                 'the same NetBIOS name: %s')
                                         % self.info['name'])

        self.generate_auth(trustdom_secret)

        info = lsa.TrustDomainInfoInfoEx()
        info.domain_name.string = another_domain.info['dns_domain']
        info.netbios_name.string = another_domain.info['name']
        info.sid = security.dom_sid(another_domain.info['sid'])
        info.trust_direction = lsa.LSA_TRUST_DIRECTION_INBOUND
        if trust_type == TRUST_BIDIRECTIONAL:
            info.trust_direction |= lsa.LSA_TRUST_DIRECTION_OUTBOUND
        info.trust_type = lsa.LSA_TRUST_TYPE_UPLEVEL
        info.trust_attributes = 0
        if trust_external:
            info.trust_attributes |= lsa.LSA_TRUST_ATTRIBUTE_NON_TRANSITIVE

        try:
            dname = lsa.String()
            dname.string = another_domain.info['dns_domain']
            res = self._pipe.QueryTrustedDomainInfoByName(
                self._policy_handle,
                dname,
                lsa.LSA_TRUSTED_DOMAIN_INFO_FULL_INFO
            )
            if res.info_ex.trust_type != lsa.LSA_TRUST_TYPE_UPLEVEL:
                msg = _('There is already a trust to {ipa_domain} with '
                        'unsupported type {trust_type}. Please remove '
                        'it manually on AD DC side.')
                ttype = trust_type_string(
                    res.info_ex.trust_type, res.info_ex.trust_attributes
                )
                err = msg.format(
                    ipa_domain=another_domain.info['dns_domain'],
                    trust_type=ttype)

                raise errors.ValidationError(
                    name=_('AD domain controller'),
                    error=err
                )

            self._pipe.DeleteTrustedDomain(self._policy_handle,
                                           res.info_ex.sid)
        except RuntimeError as e:
            # pylint: disable=unbalanced-tuple-unpacking
            num, _message = e.args
            # pylint: enable=unbalanced-tuple-unpacking
            # Ignore anything but access denied (NT_STATUS_ACCESS_DENIED)
            if num == -1073741790:
                raise access_denied_error

        try:
            trustdom_handle = self._pipe.CreateTrustedDomainEx2(
                                           self._policy_handle,
                                           info, self.auth_info,
                                           security.SEC_STD_DELETE)
        except RuntimeError as e:
            raise assess_dcerpc_error(e)

        # We should use proper trustdom handle in order to modify the
        # trust settings. Samba insists this has to be done with LSA
        # OpenTrustedDomain* calls, it is not enough to have a handle
        # returned by the CreateTrustedDomainEx2 call.
        trustdom_handle = self._pipe.OpenTrustedDomainByName(
                                         self._policy_handle,
                                         dname,
                                         security.SEC_FLAG_MAXIMUM_ALLOWED)
        try:
            infocls = lsa.TrustDomainInfoSupportedEncTypes()
            infocls.enc_types = security.KERB_ENCTYPE_RC4_HMAC_MD5
            infocls.enc_types |= security.KERB_ENCTYPE_AES128_CTS_HMAC_SHA1_96
            infocls.enc_types |= security.KERB_ENCTYPE_AES256_CTS_HMAC_SHA1_96
            self._pipe.SetInformationTrustedDomain(
                           trustdom_handle,
                           lsa.LSA_TRUSTED_DOMAIN_SUPPORTED_ENCRYPTION_TYPES,
                           infocls)
        except RuntimeError as e:
            # We can ignore the error here -- changing enctypes is for
            # improved security but the trust will work with default values as
            # well. In particular, the call may fail against Windows 2003
            # server as that one doesn't support AES encryption types
            pass

        if not trust_external:
            try:
                info = self._pipe.QueryTrustedDomainInfo(
                                      trustdom_handle,
                                      lsa.LSA_TRUSTED_DOMAIN_INFO_INFO_EX)
                info.trust_attributes |= lsa.LSA_TRUST_ATTRIBUTE_FOREST_TRANSITIVE
                self._pipe.SetInformationTrustedDomain(
                                      trustdom_handle,
                                      lsa.LSA_TRUSTED_DOMAIN_INFO_INFO_EX, info)
            except RuntimeError as e:
                logger.error(
                      'unable to set trust transitivity status: %s', str(e))

        # Updating forest trust info may fail
        # If it failed due to topology conflict, it may be fixed automatically
        # update_ftinfo() will through exceptions in that case
        # Note that MS-LSAD 3.1.4.7.16 says:
        # -------------------------
        # The server MUST also make sure that the trust attributes associated
        # with the trusted domain object referenced by the TrustedDomainName
        # parameter has the TRUST_ATTRIBUTE_FOREST_TRANSITIVE set.
        # If the attribute is not present, the server MUST return
        # STATUS_INVALID_PARAMETER.
        # -------------------------
        # Thus, we must not update forest trust info for the external trust
        if self.info['is_pdc'] and not trust_external:
            self.update_ftinfo(another_domain)

    def verify_trust(self, another_domain):
        def retrieve_netlogon_info_2(logon_server, domain, function_code, data):
            try:
                netr_pipe = netlogon.netlogon(domain.binding,
                                              domain.parm, domain.creds)
                result = netr_pipe.netr_LogonControl2Ex(
                                           logon_server=logon_server,
                                           function_code=function_code,
                                           level=2,
                                           data=data)
                return result
            except RuntimeError as e:
                raise assess_dcerpc_error(e)

        result = retrieve_netlogon_info_2(None, self,
                                          netlogon.NETLOGON_CONTROL_TC_VERIFY,
                                          another_domain.info['dns_domain'])

        if result and result.flags and netlogon.NETLOGON_VERIFY_STATUS_RETURNED:
            if result.pdc_connection_status[0] != 0 and \
               result.tc_connection_status[0] != 0:
                if result.pdc_connection_status[1] == "WERR_ACCESS_DENIED":
                    # Most likely AD DC hit another IPA replica which
                    # yet has no trust secret replicated

                    # Sleep and repeat again
                    self.validation_attempts += 1
                    if self.validation_attempts < 10:
                        sleep(5)
                        return self.verify_trust(another_domain)

                    # If we get here, we already failed 10 times
                    srv_record_templates = (
                        '_ldap._tcp.%s',
                        '_ldap._tcp.Default-First-Site-Name._sites.dc._msdcs.%s'
                    )

                    srv_records = ', '.join(
                        [srv_record % api.env.domain
                         for srv_record in srv_record_templates]
                    )

                    error_message = _(
                        'IPA master denied trust validation requests from AD '
                        'DC %(count)d times. Most likely AD DC contacted a '
                        'replica that has no trust information replicated '
                        'yet. Additionally, please check that AD DNS is able '
                        'to resolve %(records)s SRV records to the correct '
                        'IPA server.') % dict(count=self.validation_attempts,
                                              records=srv_records)

                    raise errors.ACIError(info=error_message)

                raise assess_dcerpc_error(result.pdc_connection_status)

            return True

        return False


def fetch_domains(api, mydomain, trustdomain, creds=None, server=None):
    def communicate(td):
        td.init_lsa_pipe(td.info['dc'])
        netr_pipe = netlogon.netlogon(td.binding, td.parm, td.creds)
        # Older FreeIPA versions used netr_DsrEnumerateDomainTrusts call
        # but it doesn't provide information about non-domain UPNs associated
        # with the forest, thus we have to use netr_DsRGetForestTrustInformation
        domains = netr_pipe.netr_DsRGetForestTrustInformation(td.info['dc'], None, 0)
        return domains

    domains = None
    domain_validator = DomainValidator(api)
    configured = domain_validator.is_configured()
    if not configured:
        return None

    td = TrustDomainInstance('')
    td.parm.set('workgroup', mydomain)
    cr = credentials.Credentials()
    cr.set_kerberos_state(credentials.DONT_USE_KERBEROS)
    cr.guess(td.parm)
    cr.set_anonymous()
    cr.set_workstation(domain_validator.flatname)
    netrc = net.Net(creds=cr, lp=td.parm)
    try:
        if server:
            result = netrc.finddc(address=server,
                                  flags=nbt.NBT_SERVER_LDAP | nbt.NBT_SERVER_DS)
        else:
            result = netrc.finddc(domain=trustdomain,
                                  flags=nbt.NBT_SERVER_LDAP | nbt.NBT_SERVER_DS)
    except RuntimeError as e:
        raise assess_dcerpc_error(e)

    td.info['dc'] = unicode(result.pdc_dns_name)
    td.info['name'] = unicode(result.dns_domain)
    if type(creds) is bool:
        # Rely on existing Kerberos credentials in the environment
        td.creds = credentials.Credentials()
        td.creds.set_kerberos_state(credentials.MUST_USE_KERBEROS)
        td.creds.guess(td.parm)
        td.creds.set_workstation(domain_validator.flatname)
        domains = communicate(td)
    else:
        # Attempt to authenticate as HTTP/ipa.master and use cross-forest trust
        # or as passed-in user in case of a one-way trust
        domval = DomainValidator(api)
        ccache_name = None
        if creds:
            domval._admin_creds = creds
            ccache_name, _principal = domval.kinit_as_administrator(
                trustdomain)
        else:
            raise errors.ValidationError(name=_('Credentials'),
                                         error=_('Missing credentials for '
                                                 'cross-forest communication'))
        td.creds = credentials.Credentials()
        td.creds.set_kerberos_state(credentials.MUST_USE_KERBEROS)
        if ccache_name:
            with ipautil.private_ccache(path=ccache_name):
                td.creds.guess(td.parm)
                td.creds.set_workstation(domain_validator.flatname)
                domains = communicate(td)

    if domains is None:
        return None

    result = {'domains': {}, 'suffixes': {}}
    # netr_DsRGetForestTrustInformation returns two types of entries:
    # domain information  -- name, NetBIOS name, SID of the domain
    # top level name info -- a name suffix associated with the forest
    # We should ignore forest root name/name suffix as it is already part
    # of trust information for IPA purposes and only add what's inside the forest
    for t in domains.entries:
        if t.type == lsa.LSA_FOREST_TRUST_DOMAIN_INFO:
            tname = unicode(t.forest_trust_data.dns_domain_name.string)
            if tname == trustdomain:
                continue
            result['domains'][tname] = {
                'cn': tname,
                'ipantflatname': unicode(
                    t.forest_trust_data.netbios_domain_name.string),
                'ipanttrusteddomainsid': unicode(
                    t.forest_trust_data.domain_sid)
            }
        elif t.type == lsa.LSA_FOREST_TRUST_TOP_LEVEL_NAME:
            tname = unicode(t.forest_trust_data.string)
            if tname == trustdomain:
                continue

            result['suffixes'][tname] = {'cn': tname}
    return result


def retrieve_remote_domain(hostname, local_flatname,
                           realm, realm_server=None,
                           realm_admin=None, realm_passwd=None):
    def get_instance(local_flatname):
        # Fetch data from foreign domain using password only
        rd = TrustDomainInstance('')
        rd.parm.set('workgroup', local_flatname)
        rd.creds = credentials.Credentials()
        rd.creds.set_kerberos_state(credentials.DONT_USE_KERBEROS)
        rd.creds.guess(rd.parm)
        return rd

    rd = get_instance(local_flatname)
    rd.creds.set_anonymous()
    rd.creds.set_workstation(hostname)
    if realm_server is None:
        rd.retrieve_anonymously(realm, discover_srv=True, search_pdc=True)
    else:
        rd.retrieve_anonymously(realm_server,
                                discover_srv=False, search_pdc=True)
    rd.read_only = True
    if realm_admin and realm_passwd:
        if 'name' in rd.info:
            names = realm_admin.split('\\')
            if len(names) > 1:
                # realm admin is in DOMAIN\user format
                # strip DOMAIN part as we'll enforce the one discovered
                realm_admin = names[-1]
            auth_string = u"%s\%s%%%s" \
                          % (rd.info['name'], realm_admin, realm_passwd)
            td = get_instance(local_flatname)
            td.creds.parse_string(auth_string)
            td.creds.set_workstation(hostname)
            if realm_server is None:
                # we must have rd.info['dns_hostname'] then
                # as it is part of the anonymous discovery
                td.retrieve(rd.info['dns_hostname'])
            else:
                td.retrieve(realm_server)
            td.read_only = False
            return td

    # Otherwise, use anonymously obtained data
    return rd


class TrustDomainJoins:
    def __init__(self, api):
        self.api = api
        self.local_domain = None
        self.remote_domain = None
        self.__allow_behavior = 0

        domain_validator = DomainValidator(api)
        self.configured = domain_validator.is_configured()

        if self.configured:
            self.local_flatname = domain_validator.flatname
            self.local_dn = domain_validator.dn
            self.__populate_local_domain()

    def allow_behavior(self, *flags):
        for f in flags:
            self.__allow_behavior |= int(f)

    def __populate_local_domain(self):
        # Initialize local domain info using kerberos only
        ld = TrustDomainInstance(self.local_flatname)
        ld.creds = credentials.Credentials()
        ld.creds.set_kerberos_state(credentials.MUST_USE_KERBEROS)
        ld.creds.guess(ld.parm)
        ld.creds.set_workstation(ld.hostname)
        ld.retrieve(installutils.get_fqdn())
        self.local_domain = ld

    def populate_remote_domain(self, realm, realm_server=None,
                               realm_admin=None, realm_passwd=None):
        self.remote_domain = retrieve_remote_domain(
            self.local_domain.hostname,
            self.local_domain.info['name'],
            realm,
            realm_server=realm_server,
            realm_admin=realm_admin,
            realm_passwd=realm_passwd)

    def get_realmdomains(self):
        """
        Generate list of records for forest trust information about
        our realm domains. Note that the list generated currently
        includes only top level domains, no exclusion domains, and
        no TDO objects as we handle the latter in a separate way
        """
        if self.local_domain.read_only:
            return

        self.local_domain.ftinfo_records = []

        realm_domains = self.api.Command.realmdomains_show()['result']
        # Use realmdomains' modification timestamp
        # to judge records' last update time
        entry = self.api.Backend.ldap2.get_entry(
                    realm_domains['dn'], ['modifyTimestamp'])
        # Convert the timestamp to Windows 64-bit timestamp format
        trust_timestamp = long(
                time.mktime(
                     entry.single_value.get('modifytimestamp').timetuple()
                )*1e7+116444736000000000)

        forest = DNSName(self.local_domain.info['dns_forest'])
        # tforest is IPA forest. keep the line below for future checks
        # tforest = DNSName(self.remote_domain.info['dns_forest'])
        for dom in realm_domains['associateddomain']:
            d = DNSName(dom)

            # We should skip all DNS subdomains of our forest
            # because we are going to add *.<forest> TLN anyway
            if forest.is_superdomain(d) and forest != d:
                continue

            # We also should skip single label TLDs as they
            # cannot be added as TLNs
            if len(d.labels) == 1:
                continue

            ftinfo = dict()
            ftinfo['rec_name'] = dom
            ftinfo['rec_time'] = trust_timestamp
            ftinfo['rec_type'] = lsa.LSA_FOREST_TRUST_TOP_LEVEL_NAME
            self.local_domain.ftinfo_records.append(ftinfo)

    def join_ad_full_credentials(self, realm, realm_server, realm_admin,
                                 realm_passwd, trust_type):
        if not self.configured:
            return None

        if not(isinstance(self.remote_domain, TrustDomainInstance)):
            self.populate_remote_domain(
                realm,
                realm_server,
                realm_admin,
                realm_passwd
            )

        trust_external = bool(self.__allow_behavior & TRUST_JOIN_EXTERNAL)
        if self.remote_domain.info['dns_domain'] != \
           self.remote_domain.info['dns_forest']:
            if not trust_external:
                raise errors.NotAForestRootError(
                          forest=self.remote_domain.info['dns_forest'],
                          domain=self.remote_domain.info['dns_domain'])

        if not self.remote_domain.read_only:
            trustdom_pass = samba.generate_random_password(128, 128)
            self.get_realmdomains()

            # Establishing trust may throw an exception for topology
            # conflict. If it was solved, re-establish the trust again
            # Otherwise let the CLI to display a message about the conflict
            try:
                self.remote_domain.establish_trust(self.local_domain,
                                                   trustdom_pass,
                                                   trust_type, trust_external)
            except TrustTopologyConflictSolved:
                # we solved topology conflict, retry again
                self.remote_domain.establish_trust(self.local_domain,
                                                   trustdom_pass,
                                                   trust_type, trust_external)

            # For local domain we don't set topology information
            self.local_domain.establish_trust(self.remote_domain,
                                              trustdom_pass,
                                              trust_type, trust_external)
            # if trust is inbound, we don't need to verify it because
            # AD DC will respond with WERR_NO_SUCH_DOMAIN --
            # it only does verification for outbound trusts.
            result = True
            if trust_type == TRUST_BIDIRECTIONAL:
                result = self.remote_domain.verify_trust(self.local_domain)
            return dict(
                        local=self.local_domain,
                        remote=self.remote_domain,
                        verified=result
                       )
        return None

    def join_ad_ipa_half(self, realm, realm_server, trustdom_passwd, trust_type):
        if not self.configured:
            return None

        if not(isinstance(self.remote_domain, TrustDomainInstance)):
            self.populate_remote_domain(realm, realm_server, realm_passwd=None)

        trust_external = bool(self.__allow_behavior & TRUST_JOIN_EXTERNAL)
        if self.remote_domain.info['dns_domain'] != \
           self.remote_domain.info['dns_forest']:
            if not trust_external:
                raise errors.NotAForestRootError(
                          forest=self.remote_domain.info['dns_forest'],
                          domain=self.remote_domain.info['dns_domain'])

        self.local_domain.establish_trust(self.remote_domain,
                                          trustdom_passwd,
                                          trust_type, trust_external)
        return {
            'local': self.local_domain,
            'remote': self.remote_domain,
            'verified': False,
        }



# Authors:
#   Rob Crittenden <rcritten@redhat.com>
#   Pavel Zuna <pzuna@redhat.com>
#
# Copyright (C) 2009  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import six

from ipalib import api, errors
from ipalib import Str, StrEnum, Flag
from ipalib.plugable import Registry
from .baseldap import (
    external_host_param,
    add_external_pre_callback,
    add_external_post_callback,
    remove_external_post_callback,
    LDAPObject,
    LDAPCreate,
    LDAPDelete,
    LDAPUpdate,
    LDAPSearch,
    LDAPRetrieve,
    LDAPAddMember,
    LDAPRemoveMember)
from ipalib import _, ngettext
from .hbacrule import is_all
from ipapython.dn import DN

if six.PY3:
    unicode = str

__doc__ = _("""
Netgroups

A netgroup is a group used for permission checking. It can contain both
user and host values.

EXAMPLES:

 Add a new netgroup:
   ipa netgroup-add --desc="NFS admins" admins

 Add members to the netgroup:
   ipa netgroup-add-member --users=tuser1 --users=tuser2 admins

 Remove a member from the netgroup:
   ipa netgroup-remove-member --users=tuser2 admins

 Display information about a netgroup:
   ipa netgroup-show admins

 Delete a netgroup:
   ipa netgroup-del admins
""")

register = Registry()

NETGROUP_PATTERN='^[a-zA-Z0-9_.][a-zA-Z0-9_.-]*$'
NETGROUP_PATTERN_ERRMSG='may only include letters, numbers, _, -, and .'

# according to most common use cases the netgroup pattern should fit
# also the nisdomain pattern
NISDOMAIN_PATTERN=NETGROUP_PATTERN
NISDOMAIN_PATTERN_ERRMSG=NETGROUP_PATTERN_ERRMSG

output_params = (
        Str('memberuser_user?',
            label='Member User',
        ),
        Str('memberuser_group?',
            label='Member Group',
        ),
        Str('memberhost_host?',
            label=_('Member Host'),
        ),
        Str('memberhost_hostgroup?',
            label='Member Hostgroup',
        ),
    )


@register()
class netgroup(LDAPObject):
    """
    Netgroup object.
    """
    container_dn = api.env.container_netgroup
    object_name = _('netgroup')
    object_name_plural = _('netgroups')
    object_class = ['ipaobject', 'ipaassociation', 'ipanisnetgroup']
    permission_filter_objectclasses = ['ipanisnetgroup']
    search_attributes = [
        'cn', 'description', 'memberof', 'externalhost', 'nisdomainname',
        'memberuser', 'memberhost', 'member', 'usercategory', 'hostcategory',
    ]
    default_attributes = [
        'cn', 'description', 'memberof', 'externalhost', 'nisdomainname',
        'memberuser', 'memberhost', 'member', 'memberindirect',
        'usercategory', 'hostcategory',
    ]
    uuid_attribute = 'ipauniqueid'
    rdn_attribute = 'ipauniqueid'
    attribute_members = {
        'member': ['netgroup'],
        'memberof': ['netgroup'],
        'memberindirect': ['netgroup'],
        'memberuser': ['user', 'group'],
        'memberhost': ['host', 'hostgroup'],
    }
    relationships = {
        'member': ('Member', '', 'no_'),
        'memberof': ('Member Of', 'in_', 'not_in_'),
        'memberindirect': (
            'Indirect Member', None, 'no_indirect_'
        ),
        'memberuser': ('Member', '', 'no_'),
        'memberhost': ('Member', '', 'no_'),
    }
    managed_permissions = {
        'System: Read Netgroups': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'cn', 'description', 'hostcategory', 'ipaenabledflag',
                'ipauniqueid', 'nisdomainname', 'usercategory', 'objectclass',
            },
        },
        'System: Read Netgroup Membership': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'externalhost', 'member', 'memberof', 'memberuser',
                'memberhost', 'objectclass',
            },
        },
        'System: Add Netgroups': {
            'ipapermright': {'add'},
            'replaces': [
                '(target = "ldap:///ipauniqueid=*,cn=ng,cn=alt,$SUFFIX")(version 3.0;acl "permission:Add netgroups";allow (add) groupdn = "ldap:///cn=Add netgroups,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Netgroups Administrators'},
        },
        'System: Modify Netgroup Membership': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {
                'externalhost', 'member', 'memberhost', 'memberuser'
            },
            'replaces': [
                '(targetattr = "memberhost || externalhost || memberuser || member")(target = "ldap:///ipauniqueid=*,cn=ng,cn=alt,$SUFFIX")(version 3.0;acl "permission:Modify netgroup membership";allow (write) groupdn = "ldap:///cn=Modify netgroup membership,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Netgroups Administrators'},
        },
        'System: Modify Netgroups': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {'description'},
            'replaces': [
                '(targetattr = "description")(target = "ldap:///ipauniqueid=*,cn=ng,cn=alt,$SUFFIX")(version 3.0; acl "permission:Modify netgroups";allow (write) groupdn = "ldap:///cn=Modify netgroups,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Netgroups Administrators'},
        },
        'System: Remove Netgroups': {
            'ipapermright': {'delete'},
            'replaces': [
                '(target = "ldap:///ipauniqueid=*,cn=ng,cn=alt,$SUFFIX")(version 3.0;acl "permission:Remove netgroups";allow (delete) groupdn = "ldap:///cn=Remove netgroups,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Netgroups Administrators'},
        },
        'System: Read Netgroup Compat Tree': {
            'non_object': True,
            'ipapermbindruletype': 'anonymous',
            'ipapermlocation': api.env.basedn,
            'ipapermtarget': DN('cn=ng', 'cn=compat', api.env.basedn),
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'objectclass', 'cn', 'membernisnetgroup', 'nisnetgrouptriple',
            },
        },
    }

    label = _('Netgroups')
    label_singular = _('Netgroup')

    takes_params = (
        Str('cn',
            pattern=NETGROUP_PATTERN,
            pattern_errmsg=NETGROUP_PATTERN_ERRMSG,
            cli_name='name',
            label=_('Netgroup name'),
            primary_key=True,
            normalizer=lambda value: value.lower(),
        ),
        Str('description?',
            cli_name='desc',
            label=_('Description'),
            doc=_('Netgroup description'),
        ),
        Str('nisdomainname?',
            pattern=NISDOMAIN_PATTERN,
            pattern_errmsg=NISDOMAIN_PATTERN_ERRMSG,
            cli_name='nisdomain',
            label=_('NIS domain name'),
        ),
        Str('ipauniqueid?',
            cli_name='uuid',
            label='IPA unique ID',
            doc=_('IPA unique ID'),
            flags=['no_create', 'no_update'],
        ),
        StrEnum('usercategory?',
            cli_name='usercat',
            label=_('User category'),
            doc=_('User category the rule applies to'),
            values=(u'all', ),
        ),
        StrEnum('hostcategory?',
            cli_name='hostcat',
            label=_('Host category'),
            doc=_('Host category the rule applies to'),
            values=(u'all', ),
        ),
        external_host_param,
    )

    def get_primary_key_from_dn(self, dn):
        assert isinstance(dn, DN)
        if not dn.rdns:
            return u''

        first_ava = dn.rdns[0][0]
        if first_ava[0] == self.primary_key.name:
            return unicode(first_ava[1])

        try:
            entry_attrs = self.backend.get_entry(
                dn, [self.primary_key.name]
            )
            try:
                return entry_attrs[self.primary_key.name][0]
            except (KeyError, IndexError):
                return u''
        except errors.NotFound:
            return unicode(dn)


@register()
class netgroup_add(LDAPCreate):
    __doc__ = _('Add a new netgroup.')

    has_output_params = LDAPCreate.has_output_params + output_params
    msg_summary = _('Added netgroup "%(value)s"')

    msg_collision = _(u'hostgroup with name "%s" already exists. ' \
                      u'Hostgroups and netgroups share a common namespace')

    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        assert isinstance(dn, DN)
        entry_attrs.setdefault('nisdomainname', self.api.env.domain)

        try:
            test_dn = self.obj.get_dn(keys[-1])
            netgroup = ldap.get_entry(test_dn, ['objectclass'])
            if 'mepManagedEntry' in netgroup.get('objectclass', []):
                raise errors.DuplicateEntry(message=unicode(self.msg_collision % keys[-1]))
            else:
                self.obj.handle_duplicate_entry(*keys)
        except errors.NotFound:
            pass

        try:
            # when enabled, a managed netgroup is created for every hostgroup
            # make sure that we don't create a collision if the plugin is
            # (temporarily) disabled
            api.Object['hostgroup'].get_dn_if_exists(keys[-1])
            raise errors.DuplicateEntry(message=unicode(self.msg_collision % keys[-1]))
        except errors.NotFound:
            pass

        return dn


@register()
class netgroup_del(LDAPDelete):
    __doc__ = _('Delete a netgroup.')

    msg_summary = _('Deleted netgroup "%(value)s"')



@register()
class netgroup_mod(LDAPUpdate):
    __doc__ = _('Modify a netgroup.')

    has_output_params = LDAPUpdate.has_output_params + output_params
    msg_summary = _('Modified netgroup "%(value)s"')

    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        assert isinstance(dn, DN)
        try:
            entry_attrs = ldap.get_entry(dn, attrs_list)
            dn = entry_attrs.dn
        except errors.NotFound:
            raise self.obj.handle_not_found(*keys)
        if is_all(options, 'usercategory') and 'memberuser' in entry_attrs:
            raise errors.MutuallyExclusiveError(
                reason=_("user category cannot be set to 'all' while there "
                         "are allowed users")
            )
        if is_all(options, 'hostcategory') and 'memberhost' in entry_attrs:
            raise errors.MutuallyExclusiveError(
                reason=_("host category cannot be set to 'all' while there "
                         "are allowed hosts")
            )
        return dn


@register()
class netgroup_find(LDAPSearch):
    __doc__ = _('Search for a netgroup.')

    member_attributes = ['member', 'memberuser', 'memberhost', 'memberof']
    has_output_params = LDAPSearch.has_output_params + output_params
    msg_summary = ngettext(
        '%(count)d netgroup matched', '%(count)d netgroups matched', 0
    )

    takes_options = LDAPSearch.takes_options + (
        Flag('private',
            exclude='webui',
            flags=['no_option', 'no_output'],
        ),
        Flag('managed',
            cli_name='managed',
            doc=_('search for managed groups'),
            default_from=lambda private: private,
        ),
    )

    def pre_callback(self, ldap, filter, attrs_list, base_dn, scope, *args, **options):
        assert isinstance(base_dn, DN)
        # Do not display private mepManagedEntry netgroups by default
        # If looking for managed groups, we need to omit the negation search filter

        search_kw = {}
        search_kw['objectclass'] = ['mepManagedEntry']
        if not options['managed']:
            local_filter = ldap.make_filter(search_kw, rules=ldap.MATCH_NONE)
        else:
            local_filter = ldap.make_filter(search_kw, rules=ldap.MATCH_ALL)
        filter = ldap.combine_filters((local_filter, filter), rules=ldap.MATCH_ALL)
        return (filter, base_dn, scope)


@register()
class netgroup_show(LDAPRetrieve):
    __doc__ = _('Display information about a netgroup.')

    has_output_params = LDAPRetrieve.has_output_params + output_params


@register()
class netgroup_add_member(LDAPAddMember):
    __doc__ = _('Add members to a netgroup.')

    member_attributes = ['memberuser', 'memberhost', 'member']
    has_output_params = LDAPAddMember.has_output_params + output_params

    def pre_callback(self, ldap, dn, found, not_found, *keys, **options):
        assert isinstance(dn, DN)
        return add_external_pre_callback('host', ldap, dn, keys, options)

    def post_callback(self, ldap, completed, failed, dn, entry_attrs,
                      *keys, **options):
        assert isinstance(dn, DN)
        return add_external_post_callback(ldap, dn, entry_attrs,
                                          failed=failed,
                                          completed=completed,
                                          memberattr='memberhost',
                                          membertype='host',
                                          externalattr='externalhost')


@register()
class netgroup_remove_member(LDAPRemoveMember):
    __doc__ = _('Remove members from a netgroup.')

    member_attributes = ['memberuser', 'memberhost', 'member']
    has_output_params = LDAPRemoveMember.has_output_params + output_params

    def post_callback(self, ldap, completed, failed, dn, entry_attrs,
                      *keys, **options):
        assert isinstance(dn, DN)
        return remove_external_post_callback(ldap, dn, entry_attrs,
                                             failed=failed,
                                             completed=completed,
                                             memberattr='memberhost',
                                             membertype='host',
                                             externalattr='externalhost')



# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2008  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `tests.util` module.
"""

import re

import six

from ipatests import util
from ipatests.util import raises, TYPE, VALUE, LEN, KEYS
import pytest

if six.PY3:
    unicode = str

pytestmark = pytest.mark.tier0


# pytest >= 2.10 supports yield based fixtures with pytest.fixture. In
# pytest < 2.10 pytest.yield_fixture is required. But that function
# also raises a deprecation warning in pytest >= 3.0.
PYTEST_VERSION = tuple(int(p) for p in pytest.__version__.split('.'))

if PYTEST_VERSION < (2, 10):
    yield_fixture = pytest.yield_fixture
else:
    yield_fixture = pytest.fixture


pattern_type = type(re.compile(""))


class Prop:
    def __init__(self, *ops):
        self.__ops = frozenset(ops)
        self.__prop = 'prop value'

    def __get_prop(self):
        if 'get' not in self.__ops:
            raise AttributeError('get prop')
        return self.__prop

    def __set_prop(self, value):
        if 'set' not in self.__ops:
            raise AttributeError('set prop')
        self.__prop = value

    def __del_prop(self):
        if 'del' not in self.__ops:
            raise AttributeError('del prop')
        self.__prop = None

    prop = property(__get_prop, __set_prop, __del_prop)


class test_Fuzzy:
    klass = util.Fuzzy

    def test_init(self):
        inst = self.klass()
        assert inst.regex is None
        assert inst.type is None
        assert inst.test is None
        assert inst.re is None

        inst = self.klass('(foo|bar)')
        assert inst.regex == '(foo|bar)'
        assert inst.type is unicode
        assert inst.test is None
        assert isinstance(inst.re, pattern_type)

        inst = self.klass('(foo|bar)', type=str)
        assert inst.regex == '(foo|bar)'
        assert inst.type is str
        assert inst.test is None
        assert isinstance(inst.re, pattern_type)

        t = lambda other: other > 500

        inst = self.klass(test=t)
        assert inst.regex is None
        assert inst.type is None
        assert inst.test is t
        assert inst.re is None

        inst = self.klass(type=(int, float), test=t)
        assert inst.regex is None
        assert inst.type == (int, float)
        assert inst.test is t
        assert inst.re is None

    def test_repr(self):
        s = 'Fuzzy(%r, %r, %r)'
        t = lambda other: 0.0 <= other <= 1.0

        inst = self.klass()
        assert repr(inst) == s % (None, None, None)

        inst = self.klass('foo')
        assert repr(inst) == s % ('foo', unicode, None)

        inst = self.klass(type=(int, float))
        assert repr(inst) == s % (None, (int, float), None)

        inst = self.klass(type=(int, float), test=t)
        assert repr(inst) == s % (None, (int, float), t)

        inst = self.klass(test=t)
        assert repr(inst) == s % (None, None, t)

    def test_eq(self):
        assert (self.klass('bar') == u'foobar') is True
        assert (self.klass('^bar') == u'foobar') is False
        assert (self.klass('bar', type=bytes) == u'foobar') is False

        assert ('18' == self.klass()) is True
        assert ('18' == self.klass(type=int)) is False
        assert (18 == self.klass(type=int)) is True
        assert ('18' == self.klass(type=(int, str))) is True

        assert (self.klass() == '18') is True
        assert (self.klass(type=int) == '18') is False
        assert (self.klass(type=int) == 18) is True
        assert (self.klass(type=(int, str)) == '18') is True

        t = lambda other: other.endswith('bar')
        assert (self.klass(test=t) == 'foobar') is True
        assert (self.klass(test=t, type=unicode) == b'foobar') is False
        assert (self.klass(test=t) == 'barfoo') is False

        assert (False == self.klass()) is True
        assert (True == self.klass()) is True
        assert (None == self.klass()) is True


def test_assert_deepequal():
    f = util.assert_deepequal
    try:  # pylint: disable=no-member
        pretty = pytest.config.getoption("pretty_print")
    except (AttributeError, ValueError):
        pretty = False

    # LEN and KEYS formats use special function to pretty print structures
    # depending on a pytest environment settings
    def got_str(s):
        return util.struct_to_string(s, util.GOT_LEN) if pretty else str(s)

    def exp_str(s):
        ret = util.struct_to_string(s, util.EXPECTED_LEN) if pretty else str(s)
        return ret

    # Test with good scalar values:
    f(u'hello', u'hello')
    f(util.Fuzzy(), u'hello')
    f(util.Fuzzy(type=unicode), u'hello')
    f(util.Fuzzy('ell'), u'hello')
    f(util.Fuzzy(test=lambda other: other.endswith('llo')), u'hello')
    f(18, 18)
    f(util.Fuzzy(), 18)
    f(util.Fuzzy(type=int), 18)
    f(util.Fuzzy(type=(int, float), test=lambda other: other > 17.9), 18)

    # Test with bad scalar values:
    e = raises(AssertionError, f, u'hello', u'world', 'foo')
    assert str(e) == VALUE % (
        'foo', u'hello', u'world', tuple()
    )

    e = raises(AssertionError, f, b'hello', u'hello', 'foo')
    assert str(e) == TYPE % (
        'foo', bytes, unicode, b'hello', u'hello', tuple()
    )

    e = raises(AssertionError, f, 18, 18.0, 'foo')
    assert str(e) == TYPE % (
        'foo', int, float, 18, 18.0, tuple()
    )

    # Test with good compound values:
    a = [
        u'hello',
        dict(profession=u'nurse'),
        18,
    ]
    b = [
        u'hello',
        dict(profession=u'nurse'),
        18,
    ]
    f(a, b)

    # Test with bad compound values:
    b = [
        b'hello',
        dict(profession=u'nurse'),
        18,
    ]
    e = raises(AssertionError, f, a, b, 'foo')
    assert str(e) == TYPE % (
        'foo', unicode, bytes, u'hello', b'hello', (2 if six.PY2 else 0,)
    )

    b = [
        u'hello',
        dict(profession=b'nurse'),
        18,
    ]
    e = raises(AssertionError, f, a, b, 'foo')
    assert str(e) == TYPE % (
        'foo', unicode, bytes, u'nurse', b'nurse', (1, 'profession')
    )

    b = [
        u'hello',
        dict(profession=u'nurse'),
        18.0,
    ]
    e = raises(AssertionError, f, a, b, 'foo')
    assert str(e) == TYPE % (
        'foo', int, float, 18, 18.0, (0 if six.PY2 else 2,)
    )

    # List length mismatch
    b = [
        u'hello',
        dict(profession=u'nurse'),
        18,
        19
    ]
    e = raises(AssertionError, f, a, b, 'foo')
    assert str(e) == LEN % (
        'foo', 3, 4, exp_str(a), got_str(b), tuple()
    )

    b = [
        dict(profession=u'nurse'),
        18,
    ]
    e = raises(AssertionError, f, a, b, 'foo')
    assert str(e) == LEN % (
        'foo', 3, 2, exp_str(a), got_str(b), tuple()
    )

    # Dict keys mismatch:

    # Missing
    b = [
        u'hello',
        dict(),
        18,
    ]
    e = raises(AssertionError, f, a, b, 'foo')
    assert str(e) == KEYS % (
        'foo',
        ['profession'], [],
        exp_str(dict(profession=u'nurse')), got_str(dict()),
        (1,)
    )

    # Extra
    b = [
        u'hello',
        dict(profession=u'nurse', status=u'RN'),
        18,
    ]
    e = raises(AssertionError, f, a, b, 'foo')
    assert str(e) == KEYS % (
        'foo',
        [], ['status'],
        exp_str(dict(profession=u'nurse')),
        got_str(dict(profession=u'nurse', status=u'RN')),
        (1,)
    )

    # Missing + Extra
    b = [
        u'hello',
        dict(status=u'RN'),
        18,
    ]
    e = raises(AssertionError, f, a, b, 'foo')
    assert str(e) == KEYS % (
        'foo',
        ['profession'], ['status'],
        exp_str(dict(profession=u'nurse')), got_str(dict(status=u'RN')),
        (1,)
    )


def test_yes_raised():
    f = util.raises

    class SomeError(Exception):
        pass

    class AnotherError(Exception):
        pass

    def callback1():
        'raises correct exception'
        raise SomeError()

    def callback2():
        'raises wrong exception'
        raise AnotherError()

    def callback3():
        'raises no exception'

    f(SomeError, callback1)

    raised = False
    try:
        f(SomeError, callback2)
    except AnotherError:
        raised = True
    assert raised

    raised = False
    try:
        f(SomeError, callback3)
    except util.ExceptionNotRaised:
        raised = True
    assert raised


def test_no_set():
    # Tests that it works when prop cannot be set:
    util.no_set(Prop('get', 'del'), 'prop')

    # Tests that ExceptionNotRaised is raised when prop *can* be set:
    raised = False
    try:
        util.no_set(Prop('set'), 'prop')
    except util.ExceptionNotRaised:
        raised = True
    assert raised


def test_no_del():
    # Tests that it works when prop cannot be deleted:
    util.no_del(Prop('get', 'set'), 'prop')

    # Tests that ExceptionNotRaised is raised when prop *can* be set:
    raised = False
    try:
        util.no_del(Prop('del'), 'prop')
    except util.ExceptionNotRaised:
        raised = True
    assert raised


def test_read_only():
    # Test that it works when prop is read only:
    assert util.read_only(Prop('get'), 'prop') == 'prop value'

    # Test that ExceptionNotRaised is raised when prop can be set:
    raised = False
    try:
        util.read_only(Prop('get', 'set'), 'prop')
    except util.ExceptionNotRaised:
        raised = True
    assert raised

    # Test that ExceptionNotRaised is raised when prop can be deleted:
    raised = False
    try:
        util.read_only(Prop('get', 'del'), 'prop')
    except util.ExceptionNotRaised:
        raised = True
    assert raised

    # Test that ExceptionNotRaised is raised when prop can be both set and
    # deleted:
    raised = False
    try:
        util.read_only(Prop('get', 'del'), 'prop')
    except util.ExceptionNotRaised:
        raised = True
    assert raised

    # Test that AttributeError is raised when prop can't be read:
    raised = False
    try:
        util.read_only(Prop(), 'prop')
    except AttributeError:
        raised = True
    assert raised



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#

from __future__ import print_function, absolute_import

import os
import logging
from cryptography.hazmat.backends import default_backend
from cryptography import x509

from ipalib import api
from ipalib.errors import NetworkError
from ipaplatform.paths import paths
from ipapython.admintool import AdminTool
from ipaserver.install import cainstance
from ipaserver.install import installutils

logger = logging.getLogger(__name__)


class CRLGenManage(AdminTool):
    command_name = "ipa-crlgen-manage"
    usage = "%prog <enable|disable|status>"
    description = "Manage CRL Generation Master."

    def validate_options(self):
        super(CRLGenManage, self).validate_options(needs_root=True)
        installutils.check_server_configuration()

        option_parser = self.option_parser

        if not self.args:
            option_parser.error("action not specified")
        elif len(self.args) > 1:
            option_parser.error("too many arguments")

        action = self.args[0]
        if action not in {'enable', 'disable', 'status'}:
            option_parser.error("unrecognized action '{}'".format(action))

    def run(self):
        api.bootstrap(in_server=True, confdir=paths.ETC_IPA)
        api.finalize()

        try:
            api.Backend.ldap2.connect()
        except NetworkError as e:
            logger.debug("Unable to connect to the local instance: %s", e)
            raise RuntimeError("IPA must be running, please run ipactl start")
        ca = cainstance.CAInstance(api.env.realm)

        try:
            action = self.args[0]
            if action == 'enable':
                self.enable(ca)
            elif action == 'disable':
                self.disable(ca)
            elif action == 'status':
                self.status(ca)
        finally:
            api.Backend.ldap2.disconnect()

        return 0

    def check_local_ca_instance(self, raiseOnErr=False):
        if not api.Command.ca_is_enabled()['result'] or \
           not cainstance.is_ca_installed_locally():
            if raiseOnErr:
                raise RuntimeError("Dogtag CA is not installed. "
                                   "Please install a CA first with the "
                                   "`ipa-ca-install` command.")
            else:
                logger.warning(
                    "Warning: Dogtag CA is not installed on this server.")
                return False
        return True

    def enable(self, ca):
        # When the local node is not a CA, raise an Exception
        self.check_local_ca_instance(raiseOnErr=True)
        ca.setup_crlgen(True)
        logger.info("CRL generation enabled on the local host. "
                    "Please make sure to have only a single CRL generation "
                    "master.")

    def disable(self, ca):
        # When the local node is not a CA, nothing to do
        if not self.check_local_ca_instance():
            return
        ca.setup_crlgen(False)
        logger.info("CRL generation disabled on the local host. "
                    "Please make sure to configure CRL generation on another "
                    "master with %s enable", self.command_name)

    def status(self, ca):
        # When the local node is not a CA, return "disabled"
        if not self.check_local_ca_instance():
            print("CRL generation: disabled")
            return

        # Local node is a CA, check its configuration
        if ca.is_crlgen_enabled():
            print("CRL generation: enabled")
            try:
                crl_filename = os.path.join(paths.PKI_CA_PUBLISH_DIR,
                                            'MasterCRL.bin')
                with open(crl_filename, 'rb') as f:
                    crl = x509.load_der_x509_crl(f.read(), default_backend())
                    print("Last CRL update: {}".format(crl.last_update))
                    for ext in crl.extensions:
                        if ext.oid == x509.oid.ExtensionOID.CRL_NUMBER:
                            print("Last CRL Number: {}".format(
                                ext.value.crl_number))
            except IOError:
                logger.error("Unable to find last CRL")
        else:
            print("CRL generation: disabled")



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#
"""Helpers services in for cn=masters,cn=ipa,cn=etc
"""

from __future__ import absolute_import

import collections
import logging
import random

from ipapython.dn import DN
from ipalib import api
from ipalib import errors

logger = logging.getLogger(__name__)

# constants for ipaConfigString
CONFIGURED_SERVICE = u'configuredService'
ENABLED_SERVICE = u'enabledService'
HIDDEN_SERVICE = u'hiddenService'

# The service name as stored in cn=masters,cn=ipa,cn=etc. The values are:
# 0: systemd service name
# 1: start order for system service
# 2: LDAP server entry CN, also used as SERVICE_LIST key
service_definition = collections.namedtuple(
    "service_definition",
    "systemd_name startorder service_entry"
)

SERVICES = [
    service_definition('krb5kdc', 10, 'KDC'),
    service_definition('kadmin', 20, 'KPASSWD'),
    service_definition('named', 30, 'DNS'),
    service_definition('httpd', 40, 'HTTP'),
    service_definition('ipa-custodia', 41, 'KEYS'),
    service_definition('pki-tomcatd', 50, 'CA'),
    service_definition('pki-tomcatd', 51, 'KRA'),
    service_definition('smb', 60, 'ADTRUST'),
    service_definition('winbind', 70, 'EXTID'),
    service_definition('ipa-otpd', 80, 'OTPD'),
    service_definition('ipa-ods-exporter', 90, 'DNSKeyExporter'),
    service_definition('ods-enforcerd', 100, 'DNSSEC'),
    service_definition('ipa-dnskeysyncd', 110, 'DNSKeySync'),
]

SERVICE_LIST = {s.service_entry: s for s in SERVICES}


def find_providing_servers(svcname, conn=None, preferred_hosts=(), api=api):
    """Find servers that provide the given service.

    :param svcname: The service to find
    :param preferred_hosts: preferred servers
    :param conn: a connection to the LDAP server
    :param api: ipalib.API instance
    :return: list of host names in randomized order (possibly empty)

    Preferred servers are moved to the front of the list if and only if they
    are found as providing servers.
    """
    assert isinstance(preferred_hosts, (tuple, list))
    if svcname not in SERVICE_LIST:
        raise ValueError("Unknown service '{}'.".format(svcname))
    if conn is None:
        conn = api.Backend.ldap2

    dn = DN(api.env.container_masters, api.env.basedn)

    query_filter = conn.combine_filters(
        [
            conn.make_filter(
                {
                    'objectClass': 'ipaConfigObject',
                    'cn': svcname
                },
                rules=conn.MATCH_ALL,
            ),
            conn.make_filter(
                {
                    'ipaConfigString': [ENABLED_SERVICE, HIDDEN_SERVICE]
                },
                rules=conn.MATCH_ANY
            ),
        ],
        rules=conn.MATCH_ALL
    )

    try:
        entries, _trunc = conn.find_entries(
            filter=query_filter,
            attrs_list=['ipaConfigString'],
            base_dn=dn
        )
    except errors.NotFound:
        return []

    # DNS is case insensitive
    preferred_hosts = list(host_name.lower() for host_name in preferred_hosts)
    servers = []
    for entry in entries:
        servername = entry.dn[1].value.lower()
        cfgstrings = entry.get('ipaConfigString', [])
        # always consider enabled services
        if ENABLED_SERVICE in cfgstrings:
            servers.append(servername)
        # use hidden services on preferred hosts
        elif HIDDEN_SERVICE in cfgstrings and servername in preferred_hosts:
            servers.append(servername)
    # unique list of host names
    servers = list(set(servers))
    # shuffle the list like DNS SRV would randomize it
    random.shuffle(servers)
    # Move preferred hosts to front
    for host_name in reversed(preferred_hosts):
        try:
            servers.remove(host_name)
        except ValueError:
            # preferred server not found, log and ignore
            logger.warning(
                "Lookup failed: Preferred host %s does not provide %s.",
                host_name, svcname
            )
        else:
            servers.insert(0, host_name)
    return servers


def find_providing_server(svcname, conn=None, preferred_hosts=(), api=api):
    """Find a server that provides the given service.

    :param svcname: The service to find
    :param conn: a connection to the LDAP server
    :param host_name: the preferred server
    :param api: ipalib.API instance
    :return: the selected host name or None
    """
    servers = find_providing_servers(
        svcname, conn=conn, preferred_hosts=preferred_hosts, api=api
    )
    if not servers:
        return None
    else:
        return servers[0]


def get_masters(conn=None, api=api):
    """Get all master hostnames

    :param conn: a connection to the LDAP server
    :param api: ipalib.API instance
    :return: list of hostnames
    """
    if conn is None:
        conn = api.Backend.ldap2

    dn = DN(api.env.container_masters, api.env.basedn)
    entries = conn.get_entries(dn, conn.SCOPE_ONELEVEL, None, ['cn'])
    return list(e['cn'][0] for e in entries)


def is_service_enabled(svcname, conn=None, api=api):
    """Check if service is enabled on any master

    The check function only looks for presence of service entries. It
    ignores enabled/hidden flags.

    :param svcname: The service to find
    :param conn: a connection to the LDAP server
    :param api: ipalib.API instance
    :return: True/False
    """
    if svcname not in SERVICE_LIST:
        raise ValueError("Unknown service '{}'.".format(svcname))
    if conn is None:
        conn = api.Backend.ldap2

    dn = DN(api.env.container_masters, api.env.basedn)
    query_filter = conn.make_filter(
        {
            'objectClass': 'ipaConfigObject',
            'cn': svcname
        },
        rules='&'
    )
    try:
        conn.find_entries(
            filter=query_filter,
            attrs_list=[],
            base_dn=dn
        )
    except errors.NotFound:
        return False
    else:
        return True
